import io
import json
import time
from io import BytesIO
from typing import Callable, IO, Optional, Union

from llanguage_test_client.json_rpc import JsonArray, JsonObject


def duration(start_time: float) -> float:
    return time.perf_counter() - start_time


class LspJsonStream:
    istream: IO[bytes]
    timeout_s: float

    bs: Optional[bytes]
    bi: int
    message_buf: BytesIO
    num_bytes: int
    has_content_length: bool
    position: int
    sleep_s: float = 0.100
    check_server: Callable[[], bool]

    def __init__(self, istream: IO[bytes], timeout_s: float, check_server: Callable[[], bool]) -> None:
        self.istream = istream
        self.timeout_s = timeout_s
        self.bs = None
        self.bi = 0
        self.message_buf = BytesIO()
        self.check_server = check_server

    def message(self) -> str:
        return self.message_buf.getvalue().decode("utf-8")

    def next_byte(self) -> bytes:
        start_time = time.perf_counter()
        bi = self.bi
        bs = self.bs
        if (bs is not None) and (bi < len(bs)):
            bj = bi + 1
            b = bs[bi:bj]
            self.bi = bj
            self.message_buf.write(b)
            self.position += 1
            return b
        while self.check_server() \
              and ((duration(start_time) < self.timeout_s)
                   or (self.timeout_s == 0.0)):
            try:
                self.bs = self.istream.read(4096)
                if (self.bs is not None) and (len(self.bs) > 0):
                    self.bi = 0
                    return self.next_byte()
            except BlockingIOError:
                # Try again ...
                pass
            time.sleep(self.sleep_s)
        raise RuntimeError(
            f'Timed-out after {self.timeout_s} seconds while reading from the stream:\n{self.message()}'
        )

    def next(self) -> Union[JsonObject, JsonArray]:
        self.num_bytes = 0
        self.position = 0
        self.message_buf.seek(0)
        self.message_buf.truncate(0)
        return self.parse_header_name()

    def escape(self, bs: bytes) -> bytes:
        match bs:
            case b'\n':
                return b'\\n'
            case b'\t':
                return b'\\t'
            case b'\b':
                return b'\\b'
            case b'\r':
                return b'\\r'
            case b'\f':
                return b'\\f'
            case _:
                return bs

    def parse_header_name(self) -> Union[JsonObject, JsonArray]:
        start: int = self.position
        while True:
            match self.next_byte():
                case b'\r':
                    length: int = self.position - start - 1
                    b = self.next_byte()
                    if b != b'\n':
                        raise RuntimeError(
                            f"Expected \\r to be followed by \\n, not '{self.escape(b)}':\n{self.message()}"
                        )
                    if length == 0 and self.has_content_length:
                        return self.parse_body()
                    raise RuntimeError(
                        f"Reached out-of-sequence newline while parsing header name:\n{self.message()}"
                    )
                case b'\n':
                    raise RuntimeError(
                        f"Reached out-of-sequence newline while parsing header name:\n{self.message()}"
                    )
                case b':':
                    length: int = self.position - start - 1
                    self.message_buf.seek(start)
                    header_name: str = self.message_buf.read(length).upper().decode("utf-8")
                    self.message_buf.seek(0, io.SEEK_END)
                    self.has_content_length = (header_name == "CONTENT-LENGTH")
                    return self.parse_header_value()
                case _:
                    pass

    def parse_header_value(self) -> Union[JsonObject, JsonArray]:
        start: int = self.position
        b = self.next_byte()
        while (b == b' ') or (b == b'\t'):
            b = self.next_byte()
        while True:
            match b:
                case b'\r':
                    length: int = self.position - start - 1
                    b = self.next_byte()
                    if b != b'\n':
                        raise RuntimeError(
                            f"Expected \\r to be followed by \\n, not '{self.escape(b)}':\n{self.message()}"
                        )
                    if self.has_content_length:
                        self.message_buf.seek(start)
                        header_value: str = self.message_buf.read(length).decode("utf-8")
                        self.message_buf.seek(0, io.SEEK_END)
                        self.num_bytes = int(header_value)
                    return self.parse_header_name()
                case b'\n':
                    raise RuntimeError(
                        f"Reached out-of-sequence newline while parsing header name:\n{self.message()}"
                    )
                case _:
                    b = self.next_byte()

    def parse_body(self) -> Union[JsonObject, JsonArray]:
        remaining_bytes = self.num_bytes
        bs = self.bs
        bi = self.bi
        if (bs is not None) and (bi < len(bs)):
            bk = len(bs) - bi
            bj = bi + min(bk, remaining_bytes)
            self.message_buf.write(bs[bi:bj])
            remaining_bytes -= bk
            self.bi = bj
        start_time = time.perf_counter()
        while (remaining_bytes > 0) \
              and self.check_server() \
              and ((duration(start_time) < self.timeout_s)
                   or (self.timeout_s == 0.0)):
            try:
                bs = self.istream.read(remaining_bytes)
                if bs is not None and len(bs) > 0:
                    self.message_buf.write(bs)
                    remaining_bytes -= len(bs)
                    continue
            except BlockingIOError:
                pass
            time.sleep(self.sleep_s)
        if remaining_bytes > 0:
            raise RuntimeError(
                f'Timed-out after {self.timeout_s} seconds while reading from the stream:\n{self.message()}'
            )
        self.message_buf.seek(self.message_buf.tell() - self.num_bytes)
        body: str = self.message_buf.read().decode("utf-8")
        self.message_buf.seek(0, io.SEEK_END)
        return json.loads(body)



import unittest
import jupyter_kernel_test as jkt

class IRKernelTests(jkt.KernelTests):
    kernel_name = "fortran"

    language_name = "fortran"

    file_extension = ".f90"

    code_hello_world = 'print *, "hello, world"'
    #code_stderr = "1x"

    #complete_code_samples = ['1', 'print *, "hello, world"',
    #        "integer :: i"]
    #incomplete_code_samples = ["subroutine f("]

    #code_generate_error = "1x"

    code_execute_result = [
        {'code': "1+2+3", 'result': "6"},
        {'code': "1+2", 'result': "3"},
        {'code': "integer :: x; x = 5; x*2", 'result': "10"},
    ]

if __name__ == '__main__':
    unittest.main()



import re
from functools import wraps
from typing import Any, Dict, List, Set, Tuple

RE_CAMEL_CASE_TO_UNDERSCORE: re.Pattern = \
    re.compile("(?:([a-z])([A-Z])|([A-Z0-9])([A-Z][a-z]))")
RE_UPPER_FIRST: re.Pattern = \
    re.compile("^(?:([A-Z])([^A-Z])|([A-Z]+?)([A-Z][^A-Z]|$))")

def match_pair_to_underscore(match: re.Match) -> str:
    if match.group(1):
        lhs = match.group(1)
        rhs = match.group(2)
    else:
        lhs = match.group(3)
        rhs = match.group(4)
    return f"{lhs}_{rhs}"

def memoize(fn):
    memo = {}

    @wraps(fn)
    def wrapper(*args):
        if args in memo:
            return memo[args]
        value = fn(*args)
        memo[args] = value
        return value

    return wrapper

@memoize
def camel_case_to_underscore(source: str) -> str:
    target = RE_CAMEL_CASE_TO_UNDERSCORE.sub(
        match_pair_to_underscore,
        source
    )

    while source != target:
        source = target
        target = RE_CAMEL_CASE_TO_UNDERSCORE.sub(
            match_pair_to_underscore,
            source
        )

    return target.upper()

def match_pair_to_lower_first(match: re.Match) -> str:
    if match.group(1):
        lhs = match.group(1)
        rhs = match.group(2)
    else:
        lhs = match.group(3)
        rhs = match.group(4)
    return f"{lhs.lower()}{rhs}"

@memoize
def lower_first(name: str) -> str:
    if len(name) > 0:
        return RE_UPPER_FIRST.sub(
            match_pair_to_lower_first,
            name
        )
    return ""

@memoize
def upper_first(name: str) -> str:
    if name == "uinteger":
        return "UInteger"
    if len(name) > 0:
        return name[0].upper() + name[1:]
    return ""

@memoize
def method_to_camel_case(method: str) -> str:
    if method.startswith(r'$/'):
        method = method[2:]
    return "_".join(map(upper_first, method.split(r'/')))

RESERVED_NAMES: Set[str] = {
    "any",
    "array",
    "boolean",
    "decimal",
    "integer",
    "null",
    "object",
    "string",
    "uinteger",
}

@memoize
def normalize_name(name: str) -> str:
    match name:
        case "LSPAny":
            return normalize_name("any")
        case "LSPObject":
            return normalize_name("object")
        case "LSPArray":
            return normalize_name("array")
        case "int":
            return normalize_name("integer")
        case "unsigned int":
            return normalize_name("uinteger")
        case "std::string" | "DocumentUri" | "RegExp" | "URI":
            return normalize_name("string")
        case "bool":
            return normalize_name("boolean")
        case "double":
            return normalize_name("decimal")
        case "std::nullptr_t":
            return normalize_name("null")
        case _:
            return name

@memoize
def rename_enum(enum_name: str) -> str:
    enum_name = normalize_name(enum_name)
    enum_name = upper_first(enum_name)
    return enum_name

@memoize
def rename_type(old_name: str) -> str:
    new_name = old_name
    if new_name.lower() in RESERVED_NAMES:
        new_name = f"{new_name}_t"
    return new_name

@memoize
def rename_field(field_name: str) -> str:
    match field_name:
        case 'LSPAny':
            field_name = 'any'
        case 'LSPObject':
            field_name = 'object'
        case 'LSPArray':
            field_name = 'array'
    field_name = lower_first(field_name)
    return field_name

@memoize
def any_enum(type_name: str) -> str:
    match type_name:
        case "object" | "array" | "string" | "integer" | "uinteger" | "decimal" | "boolean" | "null":
            return rename_enum(type_name)
        case "URI" | "DocumentUri" | "RegExp":
            return rename_enum("string")
        case _:
            return rename_enum("object")

@memoize
def send_fn(method: str) -> str:
    if method.startswith(r'$/'):
        method = method[2:]
    fn_nym = method.replace(r'/', '_')
    fn_nym = upper_first(fn_nym)
    return f"send{fn_nym}"

@memoize
def receive_fn(method: str) -> str:
    if method.startswith(r'$/'):
        method = method[2:]
    fn_nym = method.replace(r'/', '_')
    fn_nym = upper_first(fn_nym)
    return f"receive{fn_nym}"

def as_enumeration_spec(type_name: str, enumeration: List[Tuple[str, str]]) -> Dict[str, Any]:
    return {
        "name": type_name,
        "type": {
            "kind": "base",
            "name": "string",
        },
        "values": [
            {
                "name": enum_name,
                "value": enum_value,
            }
            for enum_name, enum_value in enumeration
        ],
    }



#!/usr/bin/env python

import argparse
import subprocess as sp
import os

# Initialization
NO_OF_THREADS = 8 # default no of threads is 8
SUPPORTED_BACKENDS = ['llvm', 'llvm2', 'llvm_rtlib', 'c', 'cpp', 'x86', 'wasm',
                      'gfortran', 'flang', 'llvmImplicit', 'llvmStackArray', 'llvm_integer_8',
                      'llvm_infer', 'fortran', 'c_nopragma', 'llvm_nopragma',
                      'llvm_wasm', 'llvm_wasm_emcc', 'llvm_omp', 'llvm_submodule',
                      'mlir', 'mlir_omp', 'mlir_llvm_omp', 'llvm_goc',
                      'target_offload', 'llvm_single_invocation', 'metal', 'cuda']
SUPPORTED_STANDARDS = ['lf', 'f23', 'legacy']
BASE_DIR = os.path.dirname(os.path.realpath(__file__))
LFORTRAN_PATH = f"{BASE_DIR}/../src/bin"

fast_tests = "no"
detect_leak_tests = "no"
nofast_llvm16 = "no"
separate_compilation = "no"
use_ninja = False
user_specified_threads = False
verbose = False

def run_cmd(cmd, cwd=None):
    print(f"+ {cmd}")
    process = sp.run(cmd, shell=True, cwd=cwd)
    if process.returncode != 0:
        print("Command failed.")
        exit(1)

def run_test(backend, std, test_pattern=None):
    run_cmd(f"mkdir {BASE_DIR}/test-{backend}")
    if std == "f23":
        std_string = "-DSTD_F23=yes"
    elif std == "legacy":
        std_string = "-DSTD_LEGACY=yes"
    elif std == "lf":
        std_string = ""
    else:
        raise Exception("Unsupported standard")

    cwd=f"{BASE_DIR}/test-{backend}"

    # Skip CMake's Fortran compiler detection for lfortran, since it tries
    # `-c` which requires the LLVM backend (not available for all backends).
    if backend not in ("gfortran", "flang"):
        skip_fc_detection = ("-DCMAKE_Fortran_COMPILER_WORKS=1 "
                             "-DCMAKE_Fortran_COMPILER_FORCED=1")
    else:
        skip_fc_detection = ""

    # Conditionally use Ninja or Make (default)
    if use_ninja:
        # Use Ninja generator for faster builds
        # Set CMAKE_Fortran_PREPROCESS_SOURCE which is required by Ninja but missing for lfortran
        generator_flags = (f"-G Ninja {skip_fc_detection} "
                          "-DCMAKE_Fortran_PREPROCESS_SOURCE=\"<CMAKE_Fortran_COMPILER> <DEFINES> <INCLUDES> <FLAGS> "
                          "-E <SOURCE> -o <PREPROCESSED_SOURCE>\"")
    else:
        # Use default Make generator
        generator_flags = skip_fc_detection

    common=f" {generator_flags} -DCURRENT_BINARY_DIR={BASE_DIR}/test-{backend} -S {BASE_DIR} -B {BASE_DIR}/test-{backend}"
    if backend == "gfortran":
        run_cmd(f"FC=gfortran cmake" + common,
                cwd=cwd)
    elif backend == "flang":
        # Resolve flang to find its LLVM install prefix, then use the
        # matching clang as CC so it can find ISO_Fortran_binding.h.
        import shutil, pathlib
        flang_path = pathlib.Path(shutil.which("flang")).resolve()
        flang_prefix = flang_path.parent.parent
        clang_bin = flang_prefix / "bin" / "clang"
        run_cmd(f"FC=flang CC={clang_bin} cmake -DLFORTRAN_BACKEND=flang" + common,
                cwd=cwd)
    elif backend == "cpp":
        run_cmd(f"FC=lfortran FFLAGS=\"--openmp\" cmake -DLFORTRAN_BACKEND={backend} -DFAST={fast_tests} "
                f"-DDETECT_LEAK={detect_leak_tests} -DLLVM_GOC={separate_compilation} "
                f"-DNOFAST_LLVM16={nofast_llvm16} {std_string}" + common,
                cwd=cwd)
    elif backend == "fortran":
        run_cmd(f"FC=lfortran cmake -DLFORTRAN_BACKEND={backend} "
            f"-DFAST={fast_tests} -DDETECT_LEAK={detect_leak_tests} "
            f"-DLLVM_GOC={separate_compilation} -DNOFAST_LLVM16={nofast_llvm16} "
            f"-DCMAKE_Fortran_FLAGS=\"-fPIC\" {std_string}" + common,
                cwd=cwd)
    else:
        run_cmd(f"FC=lfortran cmake -DLFORTRAN_BACKEND={backend} -DFAST={fast_tests} "
                f"-DDETECT_LEAK={detect_leak_tests} -DLLVM_GOC={separate_compilation} "
                f"{std_string} -DNOFAST_LLVM16={nofast_llvm16} " + common,
                cwd=cwd)

    # If a test pattern is provided, find matching tests and build only those
    if test_pattern:
        # Query ctest to find which tests match the pattern
        result = sp.run(f"ctest -N -R {test_pattern}", shell=True, cwd=cwd,
                       stdout=sp.PIPE, stderr=sp.PIPE, text=True)
        if result.returncode != 0:
            print("Failed to query tests with ctest")
            exit(1)

        # Parse the output to extract test names
        # Output format: "  Test #123: test_name"
        import re
        test_names = []
        for line in result.stdout.split('\n'):
            match = re.match(r'\s+Test\s+#\d+:\s+(\S+)', line)
            if match:
                test_names.append(match.group(1))

        if not test_names:
            print(f"No tests match pattern: {test_pattern}")
            exit(1)

        print(f"Building {len(test_names)} test(s): {', '.join(test_names)}")
        # Build only the matching test targets
        build_cmd = "ninja" if use_ninja else "make"
        # Ninja uses all cores by default, so only specify -j if user provided it
        # Make needs -j specified, so use default or user-provided value
        if use_ninja and not user_specified_threads:
            # User didn't specify -j, let ninja use all cores
            j_flag = ""
        else:
            j_flag = f" -j{NO_OF_THREADS}"
        v_flag = " -v" if verbose and use_ninja else ""
        v_env = "VERBOSE=1 " if verbose and not use_ninja else ""
        for test_name in test_names:
            run_cmd(f"{v_env}{build_cmd}{j_flag}{v_flag} {test_name}", cwd=cwd)
    else:
        # Build all tests
        build_cmd = "ninja" if use_ninja else "make"
        if use_ninja and not user_specified_threads:
            j_flag = ""
        else:
            j_flag = f" -j{NO_OF_THREADS}"
        v_flag = " -v" if verbose and use_ninja else ""
        v_env = "VERBOSE=1 " if verbose and not use_ninja else ""
        run_cmd(f"{v_env}{build_cmd}{j_flag}{v_flag}", cwd=cwd)

    # Build ctest command with optional test pattern filter
    ctest_cmd = f"ctest -j{NO_OF_THREADS} --output-on-failure"
    if verbose:
        ctest_cmd += " -V"
    if test_pattern:
        ctest_cmd += f" -R {test_pattern}"
    run_cmd(ctest_cmd, cwd=cwd)


def test_backend(backend, std, test_pattern=None):
    if backend not in SUPPORTED_BACKENDS:
        raise Exception(f"Unsupported Backend: {backend}\n")
    if std not in SUPPORTED_STANDARDS:
        raise Exception(f"Unsupported Backend: {std}\n")

    run_test(backend, std, test_pattern)

def check_module_names():
    from glob import glob
    import re
    mod = re.compile(
        r'(?im)^\s*(?:module|submodule)\b\s*(?:\([^)]+\))?\s*(?!\bprocedure\b)(\w+)'
    )
    files = glob("*.f90")
    module_names = []
    file_names = []
    for file in files:
        with open(file) as f:
            content = f.read()
        s = mod.search(content)
        if s:
            module_names.append(s.group(1).lower())
            file_names.append(file)
    for i in range(len(module_names)):
        name = module_names[i]
        if name in module_names[i+1:]:
            print("FAIL: Found a duplicate module name")
            print("Name:", name)
            print("Filename:", file_names[i])
            raise Exception("Duplicate module names")
    print("OK: All module names are unique")

def get_args():
    parser = argparse.ArgumentParser(description="LFortran Integration Test Suite")
    parser.add_argument("-j", "-n", "--no_of_threads", type=int,
                help="Parallel testing on given number of threads")
    parser.add_argument("-b", "--backends", nargs="*", default=["llvm"], type=str,
                help="Test the requested backends (%s)" % \
                        ", ".join(SUPPORTED_BACKENDS))
    parser.add_argument("--std", type=str, default="lf",
                help="Run tests with the requested Fortran standard: ".join(SUPPORTED_STANDARDS))
    parser.add_argument("-f", "--fast", action='store_true',
                help="Run supported tests with --fast")
    parser.add_argument("--detect-leak", "--detect-leaks", dest="detect_leak", action='store_true',
                help="Run LFortran tests with --detect-leaks")
    parser.add_argument("-sc", "--separate_compilation", action='store_true',
                help="Run tests with --separate-compilation")
    parser.add_argument("-nf16", "--no_fast_till_llvm16", action='store_true',
                help="Don't run unsupported tests with --fast when LLVM < 17")
    parser.add_argument("-t", "--test", type=str,
                help="Run specific tests matching pattern (regex)")
    parser.add_argument("--ninja", action='store_true',
                help="Use Ninja build system instead of Make (faster builds)")
    parser.add_argument("-m", action='store_true',
                help="Check that all module names are unique")
    parser.add_argument("-v", "--verbose", action='store_true',
                help="Show compilation commands (verbose build output)")
    return parser.parse_args()

def main():
    args = get_args()

    if args.m:
        check_module_names()
        return

    # Setup
    global NO_OF_THREADS, fast_tests, detect_leak_tests, std_f23_tests, nofast_llvm16, separate_compilation, use_ninja, user_specified_threads, verbose
    local_lfortran = os.path.join(LFORTRAN_PATH, "lfortran")
    if os.path.isfile(local_lfortran):
        os.environ["PATH"] = LFORTRAN_PATH + os.pathsep + os.environ["PATH"]

    # Set environment variable for testing
    os.environ["LFORTRAN_TEST_ENV_VAR"] = "STATUS OK!"
    # delete previously created directories (if any)
    for backend in SUPPORTED_BACKENDS:
        run_cmd(f"rm -rf {BASE_DIR}/test-{backend}")

    if args.no_of_threads:
        NO_OF_THREADS = args.no_of_threads
        user_specified_threads = True
    fast_tests = "yes" if args.fast else "no"
    detect_leak_tests = "yes" if args.detect_leak else "no"
    nofast_llvm16 = "yes" if args.no_fast_till_llvm16 else "no"
    separate_compilation = "yes" if args.separate_compilation else "no"
    use_ninja = args.ninja
    verbose = args.verbose
    test_pattern = args.test
    for backend in args.backends:
        test_backend(backend, args.std, test_pattern)

if __name__ == "__main__":
    main()



import datetime
from json import load, dump
import os
import shutil
import sys

dest_dir = sys.argv[1]
version = sys.argv[2]
lfortran_commit_sha = sys.argv[3]

# Retention policy
MAX_DEV_BUILDS = 10  # Keep only the latest 10 debug builds
# Release builds have no limit (keep all)

filename = "data.json"
if os.path.exists(filename):
    d = load(open(filename))
else:
    d = {"dev": [], "release": []}

# Add new entry
entry = {
    "url": "https://lfortran.github.io/wasm_builds/%s/%s" % \
            (dest_dir, lfortran_commit_sha),
    "version": version,
    "lfortran_commit_sha": lfortran_commit_sha,
    "created": str(datetime.datetime.now())
}
d[dest_dir].insert(0, entry)

# Apply retention policy for dev builds
if dest_dir == "dev" and len(d[dest_dir]) > MAX_DEV_BUILDS:
    print(f"Applying retention policy: keeping only {MAX_DEV_BUILDS} latest dev builds")
    
    # Get builds to remove (everything after the first MAX_DEV_BUILDS entries)
    builds_to_remove = d[dest_dir][MAX_DEV_BUILDS:]
    
    # Remove old build directories
    for build in builds_to_remove:
        build_sha = build["lfortran_commit_sha"]
        build_path = os.path.join(dest_dir, build_sha)
        if os.path.exists(build_path):
            print(f"Removing old build directory: {build_path}")
            shutil.rmtree(build_path)
        else:
            print(f"Build directory not found (already removed?): {build_path}")
    
    # Keep only the latest MAX_DEV_BUILDS entries in JSON
    d[dest_dir] = d[dest_dir][:MAX_DEV_BUILDS]
    print(f"Kept {len(d[dest_dir])} dev builds in data.json")
elif dest_dir == "release":
    print(f"Release build: no retention limit (keeping all {len(d[dest_dir])} builds)")

print("Saving to %s." % filename)
with open(filename, "w") as f:
    dump(d, f, indent=4, ensure_ascii=False, sort_keys=True)



"""
Script for conveniently building the documentation in all supported languages.

Run this script with:

.. code::

    python3 build.py


You can also pass the language as an argument:

.. code::

    python3 build.py en de


The first language will be handled as the default language.
"""

import json
import sys
import subprocess
from pathlib import Path
from typing import List, Dict

root = Path(__file__).parent
"""
Make sure to run this script from the root of the documentation repository.
"""

template = """
<!DOCTYPE HTML>
 
<meta charset="UTF-8">
<meta http-equiv="refresh" content="1; url={0}">
 
<script>
  window.location.href = "{0}"
</script>
 
<title>Page Redirection</title>
 
If you are not redirected automatically, follow the <a href='{0}'>link</a>.
"""

all_redirects = {
    "index.html": "{0}/",
    **{
        f"{name}/index.html": f"../{{0}}/{name}/"
        for name in (
            "installation",
            "language",
            "progress",
            "design",
            "developer_tutorial",
            "ast_and_asr",
            "contributing",
        )
    },
    **{
        f"intrinsics/{name}/index.html": f"../../../{{0}}/intrinsics/{name}/"
        for name in (
            "array/allocated",
            "array/cshift",
            "array/size",
            "bit/bge",
            "bit/bgt",
            "bit/bit_size",
            "bit/ble",
            "bit/blt",
            "bit/btest",
            "bit/shiftl",
            "bit/shiftr",
            "bit/rshift",
            "character/achar",
            "character/adjustl",
            "character/adjustr",
            "character/char",
            "character/lge",
            "character/len_trim",
            "kind-type/kind",
            "mathematical/acos",
            "mathematical/acosh",
            "mathematical/asin",
            "mathematical/asinh",
            "mathematical/atan",
            "mathematical/atan2",
            "mathematical/atanh",
            "mathematical/fraction",
            "misc/command_argument_count",
            "misc/cpu_time",
            "misc/date_and_time",
            "misc/new_line",
            "numeric/abs",
            "numeric/aimag",
            "numeric/aint",
            "numeric/anint",
            "numeric/ceiling",
            "numeric/cmplx",
            "numeric/conjg",
            "numeric/digits",
            "numeric/dim",
            "numeric/epsilon",
            "numeric/erf",
            "numeric/erfc",
            "numeric/exp",
            "numeric/floor",
            "numeric/gamma",
            "numeric/mod",
        )
    },
}
"""
All redirects from the original site without language component.
"""

all_languages: List[str]
"""
List of currently supported languages, taken from ``doc/src/_static/languages.json``.

To support a new language add its `language code`_ to the list.

.. _language code: https://www.sphinx-doc.org/en/master/usage/configuration.html#intl-options
"""

with open(root / "src" / "_static" / "languages.json") as fd:
    all_languages = list(json.load(fd).values())


def build_docs(language: str) -> None:
    """
    Build the documentation for a single language.

    Parameters
    ----------
    language : str
        The language to build the documentation for.
    """

    subprocess.run(
        [
            "sphinx-build",
            "-b",
            "dirhtml",
            str(root / "src"),
            str(root / "site" / language),
            f"-Dlanguage={language}",
        ],
        cwd=root,
        check=True,
    )


def build_redirects(redirects: Dict[str, str], language: str) -> None:
    """
    Build the redirects for a single language.

    Parameters
    ----------
    redirects : Dict[str, str]
        Page redirects to build.
    language : str
        The language to build the redirects for.
    """

    for source, target in redirects.items():
        source_path = root / "site" / source
        redirect = template.format(target.format(language))
        if not source_path.parent.exists():
            source_path.parent.mkdir(parents=True)
        with open(source_path, "w", encoding="utf-8") as fd:
            fd.write(redirect)


def build_all(redirects: Dict[str, str], languages: List[str]) -> None:
    """
    Build the documentation for all languages.

    Parameters
    ----------
    redirects : Dict[str, str]
        Page redirects to build.
    languages : List[str]
        List of languages to build the documentation for.
    """

    for language in languages:
        build_docs(language)

    build_redirects(redirects, languages[0])


if __name__ == "__main__":
    languages = sys.argv[1:] if len(sys.argv) > 1 else all_languages

    build_all(all_redirects, languages)

    print()
    print("Preview the documentation using")
    print()
    print(f"    python3 -m http.server -d \"{root / 'site'}\"")
    print()



from abc import ABC, abstractmethod
from collections import deque
from contextlib import contextmanager
from pathlib import Path
from typing import Collection, Deque, Iterator, Optional

DEFAULT_INDENT_PATTERN: str = "    "
DEFAULT_COLUMN_WIDTH: int = 80

class FileGenerator(ABC):
    file_path: Path

    indent_level: int = 0
    indent_pattern: str = DEFAULT_INDENT_PATTERN
    column_width: int = DEFAULT_COLUMN_WIDTH

    def __init__(self, file_path: Path) -> None:
        self.file_path = file_path

    @contextmanager
    def indent(self, num_levels: int = 1) -> Iterator["FileGenerator"]:
        self.indent_level += num_levels
        yield self
        self.indent_level -= num_levels

    def __enter__(self):
        return self.open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def open(self):
        self.file_handle = self.file_path.open("w", encoding="utf-8")
        return self

    def close(self):
        self.file_handle.close()

    def indentation(self) -> str:
        return self.indent_level * self.indent_pattern

    def inline(
        self,
        message: Optional[str] = None,
        end: Optional[str] = None,
        indent: bool = False
    ) -> None:
        if indent:
            self.file_handle.write(self.indentation())
        if message is not None:
            self.file_handle.write(message)
        if end is not None:
            self.file_handle.write(end)

    def write(self, message: str, end: Optional[str] = "\n", indent: bool = True) -> None:
        self.inline(message=message, end=end, indent=indent)

    def newline(self) -> None:
        self.file_handle.write("\n")

    @abstractmethod
    def generate_code(self) -> None:
        raise NotImplementedError()



from pathlib import Path

from llanguage_server.cxx.lsp_file_generator import gensym_context
from llanguage_server.cxx.visitors import BaseCPlusPlusLspVisitor
from llanguage_server.lsp.datatypes import LspSpec
from llanguage_server.lsp.utils import (lower_first, method_to_camel_case,
                                        receive_fn, send_fn, upper_first)
from llanguage_server.lsp.visitors import (LspAnalysisPipeline, LspSymbol)


class CPlusPlusLspLanguageServerSourceGenerator(BaseCPlusPlusLspVisitor):

    def __init__(
            self,
            output_dir: Path,
            schema: LspSpec,
            pipeline: LspAnalysisPipeline,
            namespace: str
    ) -> None:
        specification_source = output_dir / "lsp_language_server.cpp"
        super().__init__(specification_source, schema, pipeline, namespace)

    def generate_constructor(self) -> None:
        with self.gen_constructor(
                'LspLanguageServer',
                params=[
                    'ls::MessageQueue &incomingMessages',
                    'ls::MessageQueue &outgoingMessages',
                    'lsl::Logger &logger',
                ],
                sups=[
                    ('ls::LanguageServer', [
                        'incomingMessages',
                        'outgoingMessages',
                        'logger',
                    ]),
                ],
                inits=[
                    ('logger', ['logger.having("LspLanguageServer")'])
                ],
        ):
            self.write('// empty')
        self.newline()

    def generate_dispatch_request(self) -> None:
        with self.gen_fn(
                'LspLanguageServer::dispatch',
                params=[
                    'ResponseMessage &response',
                    'RequestMessage &request',
                    'IncomingRequest method'
                ]
        ):
            with self.gen_switch('method'):
                for request_spec in self.schema["requests"]:
                    if request_spec["messageDirection"] in ["clientToServer", "both"]:
                        request_method = request_spec["method"]
                        request_name = method_to_camel_case(request_method)
                        result_name = f'{request_name}Result'
                        method_enum = method_to_camel_case(request_method)
                        with self.gen_case('IncomingRequest', method_enum):
                            params_spec = request_spec.get("params", None)
                            if params_spec is not None:
                                self.write('MessageParams &messageParams = requireMessageParams(request);')
                                self.write(f'{params_spec["name"]} requestParams =')
                                with self.indent():
                                    self.write(f'transformer.as{request_name}Params(messageParams);')
                                self.write(f'{result_name} result =')
                                with self.indent(): self.write(f'{receive_fn(request_method)}(request, requestParams);')
                                self.write(f'response.result =')
                                with self.indent():
                                    self.write(f'transformer.{lower_first(result_name)}ToAny(result);')
                            else:
                                self.write(f'{result_name} result = {receive_fn(request_method)}(request);')
                                self.write(f'response.result =')
                                with self.indent():
                                    self.write(f'transformer.{lower_first(result_name)}ToAny(result);')
                            self.gen_break()
                with self.gen_default():
                    self.gen_throw(
                        'LSP_EXCEPTION',
                        'ErrorCodes::MethodNotFound',
                        '"Unsupported request method: \\"" + request.method + "\\""'
                    )
        self.newline()

    def generate_dispatch_notification(self) -> None:
        with self.gen_fn(
                'LspLanguageServer::dispatch',
                params=[
                    'ResponseMessage &/*response*/',
                    'NotificationMessage &notification',
                    'IncomingNotification method'
                ]
        ):
            with self.gen_switch('method'):
                for notification_spec in self.schema["notifications"]:
                    if notification_spec["messageDirection"] in ["clientToServer", "both"]:
                        notification_method = notification_spec["method"]
                        notification_name = method_to_camel_case(notification_method)
                        notification_enum = method_to_camel_case(notification_method)
                        with self.gen_case('IncomingNotification', notification_enum):
                            params_spec = notification_spec.get("params", None)
                            if params_spec is not None:
                                self.gen_assign(
                                    'MessageParams &messageParams',
                                    'requireMessageParams(notification)'
                                )
                                self.write(f'{params_spec["name"]} notificationParams =')
                                with self.indent():
                                    self.write(f'transformer.as{notification_name}Params(messageParams);')
                                self.write(f'{receive_fn(notification_method)}(notification, notificationParams);')
                            else:
                                self.write(f'{receive_fn(notification_method)}(notification);')
                            self.gen_break()
                with self.gen_default():
                    with self.gen_if('notification.method.compare(0, 2, "$/") == 0', end=''):
                        self.write('// NOTE: If a server or client receives notifications starting with "$/"')
                        self.write('// it is free to ignore the notification:')
                        self.write('logger.debug()')
                        with self.indent():
                            self.write(f'<< "No handler exists for method: \\"" << notification.method << "\\""')
                            self.write('<< std::endl;')
                    with self.gen_else():
                        self.gen_throw(
                            'LSP_EXCEPTION',
                            'ErrorCodes::MethodNotFound',
                            '"Unsupported notification method: \\"" + notification.method + "\\""'
                        )
        self.newline()

    def generate_dispatch_response(self) -> None:
        with self.gen_fn(
                'LspLanguageServer::dispatch',
                params=[
                    'ResponseMessage &response',
                    'const std::string &method'
                ]
        ):
            self.write('RequestId requestId;')
            with self.gen_switch('response.id.type()'):
                with self.gen_case('ResponseIdType', 'integer'):
                    self.gen_assign('requestId', 'response.id.integer()')
                    self.gen_break()
                with self.gen_case('ResponseIdType', 'string'):
                    self.gen_assign('requestId', 'response.id.string()')
                    self.gen_break()
                with self.gen_default():
                    self.gen_throw_invalid_params(
                        '("Invalid ResponseId type: " +'
                        ' ResponseIdTypeNames.at(response.id.type()))'
                    )
            self.newline()
            self.write('OutgoingRequest request;')
            with self.gen_try():
                self.write('request = outgoingRequestByValue(method);')
            with self.gen_catch('std::invalid_argument &/*e*/'):
                self.write('goto invalidMethod;')
            self.newline()
            with self.gen_switch('request'):
                for request_spec in self.schema["requests"]:
                    if request_spec["messageDirection"] in ["serverToClient", "both"]:
                        request_method = request_spec["method"]
                        request_name = method_to_camel_case(request_method)
                        result_name = f'{request_name}Result'
                        request_enum = method_to_camel_case(request_method)
                        with self.gen_case('OutgoingRequest', request_enum):
                            self.write('auto readLock = LSP_READ_LOCK(requestMutex, "requests");')
                            self.gen_assign('auto iter', 'requestsById.find(requestId.integer())')
                            with self.gen_if('iter == requestsById.end()'):
                                self.write('logger.error()')
                                with self.indent():
                                    self.write('<< "Cannot locate request with id=" << requestId.integer()')
                                    self.write('<< std::endl;')
                                self.write('return;')
                            self.gen_assign('std::shared_ptr<RequestMessage> request', 'iter->second')
                            self.write('readLock.unlock();')
                            result_spec = request_spec.get("result", None)
                            if result_spec is not None:
                                with self.gen_if('!response.result.has_value()'):
                                    self.write(f'logger.error()')
                                    with self.indent():
                                        self.write(f'<< ("Missing required attribute for method "')
                                        self.write(f'    "\\"{request_method}\\": result")')
                                        self.write('<< std::endl;')
                                    self.write('return;')
                                self.gen_assign(
                                    'LSPAny &result',
                                    f'response.result.value()'
                                )
                                self.gen_assign(
                                    f'{result_name} params',
                                    f'transformer.anyTo{upper_first(result_name)}(result)'
                                )
                                self.gen_call(f'{receive_fn(request_method)}', '*request', 'response', 'params')
                            else:
                                self.gen_call(f'{receive_fn(request_method)}', 'request', 'response')
                            self.write('auto writeLock = LSP_WRITE_LOCK(requestMutex, "requests");')
                            self.gen_assign('iter', 'requestsById.find(requestId.integer())')
                            with self.gen_if('iter != requestsById.end()'):
                                self.gen_call('requestsById.erase', 'iter')
                            self.gen_break()
                with self.gen_default():
                    self.write('invalidMethod:')
                    self.write(f'logger.error()')
                    with self.indent():
                        self.write('<< "Unsupported request method: \\""')
                        self.write('<< method << "\\"";')
        self.newline()

    def generate_require_request_params(self) -> None:
        with self.gen_fn(
                'LspLanguageServer::requireMessageParams',
                'MessageParams &',
                params=[
                    'RequestMessage &request'
                ],
                specs='const'
        ):
            with self.gen_if('request.params.has_value()'):
                self.write('return request.params.value();')
            self.gen_throw_invalid_params(
                '("RequestMessage.params must be defined for "',
                ' "method=\\"" + request.method + "\\"")'
            )
        self.newline()

    def generate_require_notification_params(self) -> None:
        with self.gen_fn(
                'LspLanguageServer::requireMessageParams',
                'MessageParams &',
                params=[
                    'NotificationMessage &notification',
                ],
                specs='const'
        ):
            with self.gen_if('notification.params.has_value()'):
                self.gen_return('notification.params.value()')
            self.gen_throw_invalid_params(
                '"NotificationMessage.params must be defined for method=\\"" + notification.method + "\\""'
            )
        self.newline()

    def generate_receive_request(self, request_spec: LspSpec) -> None:
        request_method = request_spec["method"]
        request_name = method_to_camel_case(request_method)
        result_name = f'{request_name}Result'
        params = [
            'const RequestMessage &/*request*/',
        ]
        params_spec = request_spec.get("params", None)
        if params_spec is not None:
            params.append(f'{params_spec["name"]} &/*params*/')
        receive_request = receive_fn(request_method)
        self.write(f'// request: "{request_method}"')
        with self.gen_fn(
                f'LspLanguageServer::{receive_request}',
                result_name,
                params=params
        ):
            self.gen_throw(
                'LSP_EXCEPTION',
                'ErrorCodes::MethodNotFound',
                f'"No handler exists for method: \\"{request_method}\\""'
            )
        self.newline()

    def generate_receive_notification(self, notification_spec: LspSpec) -> None:
        notification_method = notification_spec["method"]
        receive_notification = receive_fn(notification_method)
        params = [
            'const NotificationMessage &/*notification*/',
        ]
        params_spec = notification_spec.get("params", None)
        if params_spec is not None:
            params.append(f'{params_spec["name"]} &/*params*/')
        self.write(f'// notification: "{notification_method}"')
        with self.gen_fn(f'LspLanguageServer::{receive_notification}', params=params):
            if notification_method.startswith("$/"):
                self.write('// NOTE: If a server or client receives notifications starting with "$/" it')
                self.write('// is free to ignore the notification:')
                self.write('logger.debug()')
                with self.indent():
                    self.write(f'<< "No handler exists for method: \\"{notification_method}\\""')
                    self.write('<< std::endl;')
            else:
                self.gen_throw(
                    'LSP_EXCEPTION',
                    'ErrorCodes::MethodNotFound',
                    f'"No handler exists for method: \\"{notification_method}\\""'
                )
        self.newline()

    @gensym_context
    def generate_send_request(self, request_spec: LspSpec) -> None:
        request_method = request_spec["method"]
        send_request = send_fn(request_method)
        params = []
        params_spec = request_spec.get("params", None)
        if params_spec is not None:
            params.append(f'{params_spec["name"]} &params')
        self.write(f'// request: "{request_method}"')
        with self.gen_fn(
                f'LspLanguageServer::{send_request}',
                'int',
                params=params
        ):
            method_name = self.gensym_init('method', f'"{request_method}"')
            if params_spec is not None:
                params_name = self.gensym_init(
                    'params',
                    'transformer.asMessageParams(params)'
                )
                self.gen_return(f'sendRequest({method_name}, {params_name})')
            else:
                self.gen_return(f'sendRequest({method_name})')
        self.newline()

    def generate_receive_response(self, request_spec: LspSpec) -> None:
        request_method = request_spec["method"]
        receive_request = receive_fn(request_method)
        request_name = method_to_camel_case(request_method)
        params = [
            'const RequestMessage &/*request*/',
            'const ResponseMessage &/*response*/',
        ]
        result_spec = request_spec.get("result", None)
        if result_spec is not None:
            result_name = f'{request_name}Result'
            if result_spec['kind'] == 'base':
                if result_spec['name'] == 'string':
                    params.append(f'{result_name} &/*params*/')
                else:
                    params.append(f'{result_name} /*params*/')
            else:
                params.append(f'{result_name} &/*params*/')
        self.write(f'// request: "{request_method}"')
        with self.gen_fn(f'LspLanguageServer::{receive_request}', params=params):
            self.write(f'logger.debug()')
            with self.indent():
                self.write(f'<< "No handler exists for method: \\"{request_method}\\""')
                self.write(f'<< std::endl;')
        self.newline()

    @gensym_context
    def generate_send_notification(self, notification_spec: LspSpec) -> None:
        notification_method = notification_spec["method"]
        send_notification = send_fn(notification_method)
        params = []
        params_spec = notification_spec.get("params", None)
        if params_spec is not None:
            params.append(f'{params_spec["name"]} &params')
        self.write(f'// notification: "{notification_method}"')
        with self.gen_fn(f'LspLanguageServer::{send_notification}', params=params):
            method_name = self.gensym_init('method', f'"{notification_method}"')
            if params_spec is not None:
                params_name = self.gensym_init(
                    'params',
                    'transformer.asMessageParams(params)'
                )
                self.gen_call('sendNotification', method_name, params_name)
            else:
                self.gen_call('sendNotification', method_name)
        self.newline()

    def generate_prepare(self) -> None:
        with self.gen_fn(
                'LspLanguageServer::prepare',
                params=[
                    'std::string &buffer',
                    'const std::string &response'
                ],
                specs='const'
        ):
            self.write('buffer.append("Content-Type: application/vscode-jsonrpc; charset=utf-8\\r\\n");')
            self.write('buffer.append("Content-Length: ");')
            self.write('buffer.append(std::to_string(response.length()));')
            self.write('buffer.append("\\r\\n");')
            self.write('buffer.append("\\r\\n");')
            self.write('buffer.append(response);')
        self.newline()

    def generate_outgoing_request_handlers(self, request_spec: LspSpec) -> None:
        self.generate_send_request(request_spec)
        self.generate_receive_response(request_spec)

    def visit_request_symbol(self, request_symbol: LspSymbol) -> None:
        request_spec = request_symbol.spec
        match request_spec["messageDirection"]:
            case "clientToServer":
                self.generate_receive_request(request_spec)
            case "serverToClient":
                self.generate_outgoing_request_handlers(request_spec)
            case "both":
                self.generate_receive_request(request_spec)
                self.generate_outgoing_request_handlers(request_spec)
            case _:
                raise ValueError(
                    f'Unsupported messageDirection: {request_spec["messageDirection"]}'
                )

    def visit_notification_symbol(self, notification_symbol: LspSymbol) -> None:
        notification_spec = notification_symbol.spec
        match notification_spec["messageDirection"]:
            case "clientToServer":
                self.generate_receive_notification(notification_spec)
            case "serverToClient":
                self.generate_send_notification(notification_spec)
            case "both":
                self.generate_receive_notification(notification_spec)
                self.generate_send_notification(notification_spec)
            case _:
                raise ValueError(
                    f'Unsupported messageDirection: {notification_spec["messageDirection"]}'
                )

    @gensym_context
    def generate_send_request_message(self) -> None:
        with self.gen_fn(
            'LspLanguageServer::send',
            params=[
                'const RequestMessage &request',
            ],
        ):
            request_id_name = self.gensym_init('requestId', 'request.id.integer()')
            any_name = self.gensym_init(
                'any',
                'transformer.requestMessageToAny(request)'
            )
            message_name = self.gensym_init(
                'message',
                f'serializer.serialize({any_name})'
            )
            with self.block():
                self.write('auto writeLock = LSP_WRITE_LOCK(requestMutex, "requests");')
                iter_name = self.gensym_init('iter', f'requestsById.find({request_id_name})')
                with self.gen_if_eq(iter_name, 'requestsById.end()'):
                    self.gen_call(
                        'requestsById.emplace_hint',
                        iter_name,
                        request_id_name,
                        'std::make_shared<RequestMessage>(std::move(request))'
                    )
            self.gen_call('ls::LanguageServer::send', message_name)
        self.newline()

    @gensym_context
    def generate_build_request(self) -> None:
        with self.gen_fn(
            'LspLanguageServer::buildRequest',
            'RequestMessage',
            params=[
                'const std::string &method',
            ],
        ):
            request_name = self.gensym_decl('RequestMessage', 'request')
            self.gen_assign(f'{request_name}.jsonrpc', 'JSON_RPC_VERSION')
            self.gen_assign(f'{request_name}.method', 'method')
            request_id_name = self.gensym_init('requestId', 'nextRequestId()')
            self.gen_assign(f'{request_name}.id', request_id_name)
            self.gen_return(request_name)
        self.newline()

    @gensym_context
    def generate_send_custom_request(self) -> None:
        with self.gen_fn(
            'LspLanguageServer::sendRequest',
            'int',
            params=[
                'const std::string &method',
                'MessageParams &params',
            ],
        ):
            request_name = self.gensym_init('request', 'buildRequest(method)')
            self.gen_assign(f'{request_name}.params', 'std::move(params)')
            self.gen_call('send', request_name)
            self.gen_return(f'{request_name}.id.integer()')
        self.newline()

    @gensym_context
    def generate_send_custom_request_no_params(self) -> None:
        with self.gen_fn(
            'LspLanguageServer::sendRequest',
            'int',
            params=[
                'const std::string &method',
            ],
        ):
            request_name = self.gensym_init('request', 'buildRequest(method)')
            self.gen_call('send', request_name)
            self.gen_return(f'{request_name}.id.integer()')
        self.newline()

    @gensym_context
    def generate_send_notification_message(self) -> None:
        with self.gen_fn(
            'LspLanguageServer::send',
            params=[
                'const NotificationMessage &notification',
            ],
        ):
            any_name = self.gensym_init(
                'any',
                'transformer.notificationMessageToAny(notification)'
            )
            message_name = self.gensym_init(
                'message',
                f'serializer.serialize({any_name})'
            )
            self.gen_call('ls::LanguageServer::send', message_name)
        self.newline()

    @gensym_context
    def generate_build_notification(self) -> None:
        with self.gen_fn(
            'LspLanguageServer::buildNotification',
            'NotificationMessage',
            params=[
                'const std::string &method',
            ],
        ):
            notification_name = self.gensym_decl(
                'NotificationMessage',
                'notification'
            )
            self.gen_assign(f'{notification_name}.jsonrpc', 'JSON_RPC_VERSION')
            self.gen_assign(f'{notification_name}.method', 'method')
            self.gen_return(notification_name)
        self.newline()

    @gensym_context
    def generate_send_custom_notification(self) -> None:
        with self.gen_fn(
            'LspLanguageServer::sendNotification',
            params=[
                'const std::string &method',
                'MessageParams &params',
            ],
        ):
            notification_name = self.gensym_init(
                'notification',
                'buildNotification(method)'
            )
            self.gen_assign(f'{notification_name}.params', 'std::move(params)')
            self.gen_call('send', notification_name)
        self.newline()

    @gensym_context
    def generate_send_custom_notification_no_params(self) -> None:
        with self.gen_fn(
            'LspLanguageServer::sendNotification',
            params=[
                'const std::string &method',
            ],
        ):
            notification_name = self.gensym_init(
                'notification',
                'buildNotification(method)'
            )
            self.gen_call('send', notification_name)
        self.newline()

    @gensym_context
    def generate_next_request_id(self) -> None:
        with self.gen_fn('LspLanguageServer::nextRequestId', 'int'):
            self.gen_return('serialRequestId++')
        self.newline()

    def generate_code(self) -> None:
        print(f'Generating: {self.file_path}')
        self.generate_disclaimer()
        self.gen_include('cctype')
        self.gen_include('iostream')
        self.gen_include('stdexcept')
        self.newline()
        self.gen_include('server/lsp_exception.h')
        self.gen_include('server/lsp_json_parser.h')
        self.gen_include('server/lsp_language_server.h')
        self.gen_include('server/lsp_specification.h')
        self.newline()
        with self.gen_namespace(self.namespace):
            self.newline()
            self.generate_constructor()
            self.generate_next_request_id()
            self.generate_send_request_message()
            self.generate_build_request()
            self.generate_send_custom_request()
            self.generate_send_custom_request_no_params()
            self.generate_send_notification_message()
            self.generate_build_notification()
            self.generate_send_custom_notification()
            self.generate_send_custom_notification_no_params()
            self.generate_dispatch_request()
            self.generate_dispatch_notification()
            self.generate_dispatch_response()
            self.generate_require_request_params()
            self.generate_require_notification_params()
            super().generate_code()



"""
Install Python wrappers to Clang:

    conda install -c conda-forge python-clang
"""
from clang.cindex import Index, CursorKind, TypeKind

import sys
sys.path.append("../..")
from lfortran.asr import asr
from lfortran.asr.asr_check import verify_asr
from lfortran.asr.asr_to_ast import asr_to_ast
from lfortran.ast.ast_to_src import ast_to_src
from lfortran.asr.builder import (make_translation_unit,
    translation_unit_make_module, scope_add_function, make_type_integer,
    make_type_real, type_eq, make_binop, scope_add_symbol, Scope,
    function_make_var, array_is_assumed_shape)

index = Index.create()
tu = index.parse("a.h")

def process(node):
    unit = make_translation_unit()
    module = translation_unit_make_module(unit, "cxx_wrapper")
    for child in node.get_children():
        if child.kind == CursorKind.FUNCTION_DECL:
            process_function(module.symtab, child)
        else:
            print(child.kind.name, child.spelling, child.location)
            #import IPython; IPython.embed()
    verify_asr(unit)
    return unit

def process_type(t):
    if t.kind == TypeKind.FUNCTIONPROTO:
        result = process_type(t.get_result())
        args = []
        for arg in t.argument_types():
            args.append(process_type(arg))
        return "result(%s), args(%s)" % (result, ", ".join(args))
    elif t.kind == TypeKind.POINTER:
        pointee = process_type(t.get_pointee())
        return "pointer(%s)" % pointee
    elif t.kind == TypeKind.INT:
        return "int"
    else:
        print(t.spelling)
        raise Exception("Type not supported")

def process_function(scope, node):
    types = process_type(node.type)
    arg_names = []
    for arg in node.get_arguments():
        arg_names.append(arg.spelling)
    type = make_type_integer()
    return_var = asr.VariableOld(name=node.spelling, type=type)
    f = scope_add_function(scope, node.spelling, args=arg_names,
            return_var=return_var)
    for a in f.args:
        a.intent = "in"
    f.bind = asr.Bind(lang="c", name=node.spelling)

u = process(tu.cursor)
a = asr_to_ast(u)
s = ast_to_src(a)
print(s)



import argparse
import json
from abc import ABC, abstractmethod

from llanguage_server.lsp.auxiliary_schema import AUXILIARY_SCHEMA
from llanguage_server.lsp.visitors import LspAnalysisPipeline
from llanguage_server.lsp.datatypes import LspSpec


class LspCodeGenerator(ABC):
    args: argparse.Namespace
    schema: LspSpec
    pipeline: LspAnalysisPipeline

    def __init__(self, args: argparse.Namespace) -> None:
        self.args = args

    def generate_code(self) -> None:
        print('Creating parent directories')
        self.args.output_dir.mkdir(parents=True, exist_ok=True)

        print('Loading schema')
        self.schema = json.loads(self.args.schema.read())
        for key, values in AUXILIARY_SCHEMA.items():
            self.schema[key].extend(values)

        print('Preprocessing data types')
        self.pipeline = LspAnalysisPipeline(self.schema)
        self.pipeline.run()

        print('Generating files')
        self.generate_specification()
        self.generate_transformer()
        self.generate_server()

        print('Done.')

    @abstractmethod
    def generate_specification(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def generate_transformer(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def generate_server(self) -> None:
        raise NotImplementedError()



import datetime
from json import load, dump
import os
import sys

dest_dir = sys.argv[1]
version = sys.argv[2]
lfortran_commit_sha = sys.argv[3]

filename = "data.json"
if os.path.exists(filename):
    d = load(open(filename))
else:
    d = {"dev": [], "release": []}
entry = {
    "url": "https://lfortran.github.io/tarballs/%s/lfortran-%s.tar.gz" % \
            (dest_dir, version),
    "version": version,
    "lfortran_commit_sha": lfortran_commit_sha,
    "created": str(datetime.datetime.now())
}
d[dest_dir].append(entry)
print("Saving to %s." % filename)
with open(filename, "w") as f:
    dump(d, f, indent=4, ensure_ascii=False, sort_keys=True)



from abc import ABC, abstractmethod
from collections import OrderedDict, deque
from contextlib import contextmanager
from functools import wraps
from itertools import chain
from typing import (Callable, Collection, Deque, Dict, Generic, Iterable,
                    Iterator, List, Optional, Set, TypeVar, Union)

from llanguage_server.lsp.datatypes import LspSpec, LspSymbol, LspSymbolKind
from llanguage_server.lsp.utils import method_to_camel_case


def memoize_instance_unary(fn: Callable) -> Callable:
    memo = {}
    @wraps(fn)
    def wrapper(self, arg):
        if arg in memo:
            return memo[arg]
        val = fn(self, arg)
        memo[arg] = val
        return val
    return wrapper


T = TypeVar('T')


class LspSpecVisitor(ABC, Generic[T]):

    def visit_type(self, type_spec: LspSpec) -> T:
        match type_spec["kind"]:
            case "base":
                return self.visit_base(type_spec)
            case "reference":
                return self.visit_reference(type_spec)
            case "enumeration":
                return self.visit_enumeration(type_spec)
            case "array":
                return self.visit_array(type_spec)
            case "map":
                return self.visit_map(type_spec)
            case "and":
                return self.visit_and(type_spec)
            case "or":
                return self.visit_union(type_spec)
            case "tuple":
                return self.visit_tuple(type_spec)
            case "literal":
                return self.visit_inner(type_spec)
            case "stringLiteral":
                return self.visit_string_literal(type_spec)
            case "integerLiteral":
                return self.visit_integer_literal(type_spec)
            case "booleanLiteral":
                return self.visit_boolean_literal(type_spec)
            case _:
                raise ValueError(f'Unsupported Type kind: {type_spec}')

    @abstractmethod
    def visit_base(self, base_spec: LspSpec) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_reference(self, reference_spec: LspSpec) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_enumeration(self, enum_spec: LspSpec) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_array(self, array_spec: LspSpec) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_map(self, map_spec: LspSpec) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_and(self, and_spec: LspSpec) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_union(self, union_spec: LspSpec) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_tuple(self, tuple_spec: LspSpec) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_inner(self, inner_spec: LspSpec) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_string_literal(self, string_spec: LspSpec) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_integer_literal(self, integer_spec: LspSpec) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_boolean_literal(self, boolean_spec: LspSpec) -> T:
        raise NotImplementedError


class NestedContext:
    nested_names: Deque[str]

    def __init__(self) -> None:
        self.nested_names = deque()

    @contextmanager
    def nested_names_as(self, nested_names: Collection[str]) -> Iterator[str]:
        if not isinstance(nested_names, deque):
            nested_names = deque(nested_names)
        prev_nested_names = self.nested_names
        self.nested_names = nested_names
        yield self.nested_name()
        self.nested_names = prev_nested_names

    @contextmanager
    def nested_name_as(self, next_nested_name: str) -> Iterator[str]:
        with self.nested_names_as(deque([next_nested_name])) as nested_name:
            yield nested_name

    @contextmanager
    def nest_name(self, name: str) -> Iterator[str]:
        self.nested_names.append(name)
        yield self.nested_name()
        self.nested_names.pop()

    def nested_name(self, nested_names: Optional[Collection[str]] = None) -> str:
        if nested_names is None:
            nested_names = self.nested_names
        return "_".join(nested_names)


class BaseLspSpecVisitor(LspSpecVisitor[None], NestedContext):

    def __init__(self) -> None:
        LspSpecVisitor.__init__(self)
        NestedContext.__init__(self)

    def visit_base(self, base_spec: LspSpec) -> None:
        pass

    def visit_reference(self, reference_spec: LspSpec) -> None:
        pass

    def visit_enumeration(self, enum_spec: LspSpec) -> None:
        pass

    def visit_array(self, array_spec: LspSpec) -> None:
        with self.nest_name('elem'):
            elem_spec: LspSpec = array_spec["element"]
            self.visit_type(elem_spec)

    def visit_map(self, map_spec: LspSpec) -> None:
        key_spec: LspSpec = map_spec["key"]
        value_spec: LspSpec = map_spec["value"]
        with self.nest_name('key') as key_name:
            self.visit_type(key_spec)
        with self.nest_name('value') as value_name:
            self.visit_type(value_spec)

    def visit_and(self, and_spec: LspSpec) -> None:
        item_specs: List[LspSpec] = and_spec["items"]
        for item_spec in item_specs:
            self.visit_type(item_spec)

    def visit_union(self, union_spec: LspSpec) -> None:
        item_specs: List[LspSpec] = union_spec["items"]
        for union_index, item_spec in enumerate(item_specs):
            with self.nest_name(str(union_index)) as nested_name:
                self.visit_type(item_spec)

    def visit_tuple(self, tuple_spec: LspSpec) -> None:
        tuple_items: List[LspSpec] = tuple_spec["items"]
        for tuple_index, tuple_item in enumerate(tuple_items):
            with self.nest_name(str(tuple_index)) as nested_name:
                self.visit_type(tuple_item)

    def visit_inner(self, inner_spec: LspSpec) -> None:
        value_spec: LspSpec = inner_spec["value"]
        field_specs: List[LspSpec] = value_spec["properties"]
        for field_spec in field_specs:
            self.visit_field(field_spec)

    def visit_field(self, field_spec: LspSpec) -> None:
        field_name = field_spec["name"]
        with self.nest_name(field_name) as nested_name:
            type_spec = field_spec["type"]
            self.visit_type(type_spec)

    def visit_string_literal(self, string_spec: LspSpec) -> None:
        pass

    def visit_integer_literal(self, integer_spec: LspSpec) -> None:
        pass

    def visit_boolean_literal(self, boolean_spec: LspSpec) -> None:
        pass

    def visit_structure(self, struct_spec: LspSpec) -> None:
        struct_name: str = struct_spec["name"]
        with self.nested_name_as(struct_name):
            super_specs: Optional[List[LspSpec]] = struct_spec.get("extends", None)
            if super_specs is not None:
                for super_spec in super_specs:
                    self.visit_type(super_spec)
            mixin_specs: Optional[List[LspSpec]] = struct_spec.get("mixins", None)
            if mixin_specs is not None:
                for mixin_spec in mixin_specs:
                    self.visit_type(mixin_spec)
            field_specs: List[LspSpec] = struct_spec["properties"]
            for field_spec in field_specs:
                self.visit_field(field_spec)

    def visit_type_alias(self, alias_spec: LspSpec) -> None:
        alias_name = alias_spec["name"]
        with self.nested_name_as(alias_name):
            type_spec: LspSpec = alias_spec["type"]
            self.visit_type(type_spec)

    def visit_request(self, request_spec: LspSpec) -> None:
        request_method = request_spec["method"]
        request_params: Optional[LspSpec] = request_spec.get("params", None)
        if request_params is not None:
            self.visit_type(request_params)
        request_name = method_to_camel_case(request_method)
        result_name = f'{request_name}Result'
        with self.nested_name_as(result_name):
            result_spec = request_spec["result"]
            # self.visit_type(result_spec)
            partial_result_spec = request_spec.get("partialResult", None)
            if partial_result_spec is not None:
                # self.visit_type(partial_result_spec)
                if result_spec["kind"] == partial_result_spec["kind"] == "or":
                    result_types = list(result_spec["items"])
                    for item_spec in partial_result_spec["items"]:
                        if item_spec not in result_types:
                            result_types.append(item_spec)
                elif result_spec["kind"] == "or":
                    result_types = list(result_spec["items"])
                    if partial_result_spec not in result_types:
                        result_types.append(partial_result_spec)
                elif partial_result_spec["kind"] == "or":
                    result_types = list(partial_result_spec["items"])
                    if result_spec not in result_types:
                        result_types.append(result_spec)
                else:
                    result_types = [result_spec]
                    if partial_result_spec not in result_types:
                        result_types.append(partial_result_spec)
            elif result_spec["kind"] == "or":
                result_types = result_spec["items"]
            else:
                result_types = [result_spec]
            if len(result_types) > 1:
                    union_type = {
                        "kind": "or",
                        "items": result_types
                    }
                    self.visit_union(union_type)
            else:
                alias_spec: LspSpec = {
                    "name": result_name,
                    "type": result_types[0]
                }
                self.visit_type_alias(alias_spec)

    def visit_notification(self, notification_spec: LspSpec) -> None:
        notification_params: Optional[LspSpec] = notification_spec.get("params", None)
        if notification_params is not None:
            self.visit_type(notification_params)


LspNameIndex = Dict[str, LspSymbol]
LspKindIndex = Dict[LspSymbolKind, List[LspSymbol]]
LspMessageIndex = Dict[str, List[LspSymbol]]

class LspIndexer(BaseLspSpecVisitor):
    schema: LspSpec

    name_index: LspNameIndex
    kind_index: LspKindIndex
    request_index: LspMessageIndex
    notification_index: LspMessageIndex

    def __init__(self, schema: LspSpec) -> None:
        BaseLspSpecVisitor.__init__(self)
        self.schema = schema
        self.name_index = OrderedDict()
        self.kind_index = OrderedDict()
        self.request_index = OrderedDict()
        self.notification_index = OrderedDict()

    def index_by_name(self, name: str, symbol: LspSymbol) -> None:
        self.name_index[name] = symbol

    def index_by_kind(self, kind: LspSymbolKind, symbol: LspSymbol) -> None:
        if kind in self.kind_index:
            symbols = self.kind_index[kind]
        else:
            symbols = []
            self.kind_index[kind] = symbols
        symbols.append(symbol)

    def index_by_direction(self, index: LspMessageIndex, direction: str, symbol: LspSymbol) -> None:
        match direction:
            case "both":
                self.index_by_direction(index, "clientToServer", symbol)
                self.index_by_direction(index, "serverToClient", symbol)
            case "clientToServer" | "serverToClient":
                if direction in index:
                    requests = index[direction]
                else:
                    requests = []
                    index[direction] = requests
                requests.append(symbol)
            case _:
                raise ValueError(f'Unsupported direction: "{direction}"')

    def index_request_by_direction(self, direction: str, symbol: LspSymbol) -> None:
        self.index_by_direction(self.request_index, direction, symbol)

    def index_notification_by_direction(self, direction: str, symbol: LspSymbol) -> None:
        self.index_by_direction(self.notification_index, direction, symbol)

    def topological_sort(self) -> Iterable[LspSymbol]:
        return LspTopologicalIterator(self)

    def index(self) -> None:
        self.visit_enumerations()
        self.visit_structures()
        self.visit_type_aliases()
        self.visit_requests()
        self.visit_notifications()

    def visit_enumerations(self) -> None:
        for enum_spec in self.schema["enumerations"]:
            self.visit_enumeration(enum_spec)

    def visit_enumeration(self, enum_spec: LspSpec) -> None:
        enum_name: str = enum_spec["name"]
        if self.name_index.get(enum_name, None) != enum_name:
            enum_kind = LspSymbolKind.ENUMERATION
            enum_symbol = LspSymbol(enum_name, enum_kind, enum_spec)
            self.index_by_name(enum_name, enum_symbol)
            self.index_by_kind(enum_kind, enum_symbol)
            BaseLspSpecVisitor.visit_enumeration(self, enum_spec)

    def visit_structures(self) -> None:
        for struct_spec in self.schema["structures"]:
            self.visit_structure(struct_spec)

    def visit_structure(self, struct_spec: LspSpec) -> None:
        struct_name: str = struct_spec["name"]
        if self.name_index.get(struct_name, None) != struct_name:
            struct_kind = LspSymbolKind.STRUCTURE
            struct_symbol = LspSymbol(struct_name, struct_kind, struct_spec)
            self.index_by_name(struct_name, struct_symbol)
            self.index_by_kind(struct_kind, struct_symbol)
            BaseLspSpecVisitor.visit_structure(self, struct_spec)

    def visit_base(self, base_spec: LspSpec) -> None:
        base_name: str = base_spec["name"]
        if self.name_index.get(base_name, None) != base_name:
            base_kind = LspSymbolKind.BASE
            base_symbol = LspSymbol(base_name, base_kind, base_spec)
            self.index_by_name(base_name, base_symbol)
            self.index_by_kind(base_kind, base_symbol)
            BaseLspSpecVisitor.visit_base(self, base_spec)

    def share_cycles(self, child_symbol: LspSymbol) -> None:
        if len(self.nested_names) > 1:
            parent_name = self.nested_name(list(self.nested_names)[:-1])
            parent_symbol = self.name_index[parent_name]
            if parent_symbol.cycles is None:
                parent_symbol.cycles = set()
            child_symbol.cycles = parent_symbol.cycles

    def visit_array(self, array_spec: LspSpec) -> None:
        array_name = self.nested_name()
        if array_name not in self.name_index:
            array_kind = LspSymbolKind.ARRAY
            array_symbol = LspSymbol(array_name, array_kind, array_spec)
            self.index_by_name(array_name, array_symbol)
            self.index_by_kind(array_kind, array_symbol)
            BaseLspSpecVisitor.visit_array(self, array_spec)
            self.share_cycles(array_symbol)

    def visit_map(self, map_spec: LspSpec) -> None:
        map_name = self.nested_name()
        if map_name not in self.name_index:
            map_kind = LspSymbolKind.MAP
            map_symbol = LspSymbol(map_name, map_kind, map_spec)
            self.index_by_name(map_name, map_symbol)
            self.index_by_kind(map_kind, map_symbol)
            BaseLspSpecVisitor.visit_map(self, map_spec)
            self.share_cycles(map_symbol)

    def visit_tuple(self, tuple_spec: LspSpec) -> None:
        tuple_name = self.nested_name()
        if tuple_name not in self.name_index:
            tuple_kind = LspSymbolKind.TUPLE
            tuple_symbol = LspSymbol(tuple_name, tuple_kind, tuple_spec)
            self.index_by_name(tuple_name, tuple_symbol)
            self.index_by_kind(tuple_kind, tuple_symbol)
            BaseLspSpecVisitor.visit_tuple(self, tuple_spec)

    def visit_union(self, union_spec: LspSpec) -> None:
        union_name: str = self.nested_name()
        if self.name_index.get(union_name, None) != union_name:
            union_kind = LspSymbolKind.UNION
            union_symbol = LspSymbol(union_name, union_kind, union_spec)
            self.index_by_name(union_name, union_symbol)
            self.index_by_kind(union_kind, union_symbol)
            BaseLspSpecVisitor.visit_union(self, union_spec)

    def visit_inner(self, inner_spec: LspSpec) -> None:
        nested_name: str = self.nested_name()
        if self.name_index.get(nested_name, None) != nested_name:
            nested_kind = LspSymbolKind.INNER
            inner_symbol = LspSymbol(nested_name, nested_kind, inner_spec)
            self.index_by_name(nested_name, inner_symbol)
            self.index_by_kind(nested_kind, inner_symbol)
            BaseLspSpecVisitor.visit_inner(self, inner_spec)

    def visit_type_aliases(self) -> None:
        for alias_spec in self.schema["typeAliases"]:
            self.visit_type_alias(alias_spec)

    def visit_type_alias(self, alias_spec: LspSpec) -> None:
        BaseLspSpecVisitor.visit_type_alias(self, alias_spec)
        alias_name: str = alias_spec["name"]
        if alias_name not in self.name_index:
            alias_kind = LspSymbolKind.ALIAS
            alias_symbol = LspSymbol(alias_name, alias_kind, alias_spec)
            self.index_by_name(alias_name, alias_symbol)
            self.index_by_kind(alias_kind, alias_symbol)

    def visit_requests(self) -> None:
        for request_spec in self.schema["requests"]:
            self.visit_request(request_spec)

    def visit_request(self, request_spec: LspSpec) -> None:
        request_method = request_spec["method"]
        if request_method not in self.name_index:
            request_kind = LspSymbolKind.REQUEST
            request_symbol = LspSymbol(request_method, request_kind, request_spec)
            self.index_by_name(request_method, request_symbol)
            self.index_by_kind(request_kind, request_symbol)
            request_direction = request_spec["messageDirection"]
            self.index_request_by_direction(request_direction, request_symbol)
            BaseLspSpecVisitor.visit_request(self, request_spec)

    def visit_notifications(self) -> None:
        for notification_spec in self.schema["notifications"]:
            self.visit_notification(notification_spec)

    def visit_notification(self, notification_spec: LspSpec) -> None:
        notification_method = notification_spec["method"]
        if notification_method not in self.name_index:
            notification_kind = LspSymbolKind.NOTIFICATION
            notification_symbol = LspSymbol(notification_method, notification_kind, notification_spec)
            self.index_by_name(notification_method, notification_symbol)
            self.index_by_kind(notification_kind, notification_symbol)
            notification_direction = notification_spec["messageDirection"]
            self.index_notification_by_direction(notification_direction, notification_symbol)
            BaseLspSpecVisitor.visit_notification(self, notification_spec)


class LspSymbolVisitor(ABC, Generic[T]):

    def visit_symbol(self, symbol: LspSymbol) -> T:
        match symbol.kind:
            case LspSymbolKind.ALIAS:
                return self.visit_alias_symbol(symbol)
            case LspSymbolKind.ARRAY:
                return self.visit_array_symbol(symbol)
            case LspSymbolKind.BASE:
                return self.visit_base_symbol(symbol)
            case LspSymbolKind.ENUMERATION:
                return self.visit_enumeration_symbol(symbol)
            case LspSymbolKind.INNER:
                return self.visit_inner_symbol(symbol)
            case LspSymbolKind.MAP:
                return self.visit_map_symbol(symbol)
            case LspSymbolKind.NOTIFICATION:
                return self.visit_notification_symbol(symbol)
            case LspSymbolKind.REFERENCE:
                return self.visit_reference_symbol(symbol)
            case LspSymbolKind.REQUEST:
                return self.visit_request_symbol(symbol)
            case LspSymbolKind.STRUCTURE:
                return self.visit_structure_symbol(symbol)
            case LspSymbolKind.TUPLE:
                return self.visit_tuple_symbol(symbol)
            case LspSymbolKind.UNION:
                return self.visit_union_symbol(symbol)

    @abstractmethod
    def visit_alias_symbol(self, alias_symbol: LspSymbol) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_array_symbol(self, array_symbol: LspSymbol) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_base_symbol(self, base_symbol: LspSymbol) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_enumeration_symbol(self, enum_symbol: LspSymbol) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_inner_symbol(self, inner_symbol: LspSymbol) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_map_symbol(self, map_symbol: LspSymbol) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_notification_symbol(self, notification_symbol: LspSymbol) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_reference_symbol(self, reference_symbol: LspSymbol) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_request_symbol(self, request_symbol: LspSymbol) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_structure_symbol(self, struct_symbol: LspSymbol) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_tuple_symbol(self, tuple_symbol: LspSymbol) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_union_symbol(self, union_symbol: LspSymbol) -> T:
        raise NotImplementedError


class BaseLspSymbolVisitor(LspSymbolVisitor[None], BaseLspSpecVisitor):
    indexer: LspIndexer

    def __init__(self, indexer: LspIndexer) -> None:
        LspSymbolVisitor.__init__(self)
        BaseLspSpecVisitor.__init__(self)
        self.indexer = indexer

    def visit_reference(self, reference_spec: LspSpec) -> None:
        reference_name: str = reference_spec["name"]
        with self.nested_name_as(reference_name):
            reference_symbol = self.indexer.name_index[reference_name]
            self.visit_symbol(reference_symbol)
            BaseLspSpecVisitor.visit_reference(self, reference_spec)

    def visit_enumeration(self, enum_spec: LspSpec) -> None:
        enum_name: str = enum_spec["name"]
        enum_symbol = self.indexer.name_index[enum_name]
        self.visit_symbol(enum_symbol)
        BaseLspSpecVisitor.visit_enumeration(self, enum_spec)

    def visit_structure(self, struct_spec: LspSpec) -> None:
        struct_name: str = struct_spec["name"]
        struct_symbol = self.indexer.name_index[struct_name]
        self.visit_symbol(struct_symbol)
        BaseLspSpecVisitor.visit_structure(self, struct_spec)

    def visit_base(self, base_spec: LspSpec) -> None:
        base_name: str = base_spec["name"]
        base_symbol = self.indexer.name_index[base_name]
        self.visit_symbol(base_symbol)
        BaseLspSpecVisitor.visit_base(self, base_spec)

    def visit_union(self, union_spec: LspSpec) -> None:
        union_name: str = self.nested_name()
        union_symbol = self.indexer.name_index[union_name]
        self.visit_symbol(union_symbol)
        BaseLspSpecVisitor.visit_union(self, union_spec)

    def visit_inner(self, inner_spec: LspSpec) -> None:
        nested_name: str = self.nested_name()
        inner_symbol = self.indexer.name_index[nested_name]
        self.visit_symbol(inner_symbol)
        BaseLspSpecVisitor.visit_inner(self, inner_spec)

    def visit_type_alias(self, alias_spec: LspSpec) -> None:
        alias_name: str = alias_spec["name"]
        alias_symbol = self.indexer.name_index[alias_name]
        self.visit_symbol(alias_symbol)
        BaseLspSpecVisitor.visit_type_alias(self, alias_spec)

    def visit_request(self, request_spec: LspSpec) -> None:
        request_method = request_spec["method"]
        request_symbol = self.indexer.name_index[request_method]
        self.visit_symbol(request_symbol)
        BaseLspSpecVisitor.visit_request(self, request_spec)

    def visit_notification(self, notification_spec: LspSpec) -> None:
        notification_method = notification_spec["method"]
        notification_symbol = self.indexer.name_index[notification_method]
        self.visit_symbol(notification_symbol)
        BaseLspSpecVisitor.visit_notification(self, notification_spec)

    def visit_alias_symbol(self, alias_symbol: LspSymbol) -> None:
        alias_spec = alias_symbol.spec
        self.visit_type_alias(alias_spec)

    def visit_array_symbol(self, array_symbol: LspSymbol) -> None:
        array_spec = array_symbol.spec
        self.visit_array(array_spec)

    def visit_base_symbol(self, base_symbol: LspSymbol) -> None:
        base_spec = base_symbol.spec
        self.visit_base(base_spec)

    def visit_enumeration_symbol(self, enum_symbol: LspSymbol) -> None:
        enum_spec = enum_symbol.spec
        self.visit_enumeration(enum_spec)

    def visit_inner_symbol(self, inner_symbol: LspSymbol) -> None:
        inner_spec = inner_symbol.spec
        self.visit_inner(inner_spec)

    def visit_map_symbol(self, map_symbol: LspSymbol) -> None:
        map_spec = map_symbol.spec
        self.visit_map(map_spec)

    def visit_notification_symbol(self, notification_symbol: LspSymbol) -> None:
        notification_spec = notification_symbol.spec
        self.visit_notification(notification_spec)

    def visit_reference_symbol(self, reference_symbol: LspSymbol) -> None:
        reference_spec = reference_symbol.spec
        self.visit_reference(reference_spec)

    def visit_request_symbol(self, request_symbol: LspSymbol) -> None:
        request_spec = request_symbol.spec
        self.visit_request(request_spec)

    def visit_structure_symbol(self, struct_symbol: LspSymbol) -> None:
        struct_spec = struct_symbol.spec
        self.visit_structure(struct_spec)

    def visit_tuple_symbol(self, tuple_symbol: LspSymbol) -> None:
        tuple_spec = tuple_symbol.spec
        self.visit_tuple(tuple_spec)

    def visit_union_symbol(self, union_symbol: LspSymbol) -> None:
        union_spec = union_symbol.spec
        self.visit_union(union_spec)


def dependency_visitor(fn: Callable) -> Callable:
    @wraps(fn)
    def wrapper(self, symbol: LspSymbol) -> None:
        if self.dependencies is not None \
           and symbol not in self.dependencies \
           and symbol not in self.blacklist:
            if symbol is self.context \
               and symbol not in self.cyclic_dependencies:
                self.cyclic_dependencies[symbol] = True
            self.dependencies[symbol] = True
            if symbol.dependencies is not None:
                # Use the cached set of dependencies
                for dependency in symbol.dependencies:
                    if dependency is self.context \
                       and dependency not in self.cyclic_dependencies:
                        self.cyclic_dependencies[dependency] = True
                    self.dependencies[dependency] = True
            else:
                fn(self, symbol)
    return wrapper


class LspDependencyExtractor(BaseLspSymbolVisitor):
    cyclic_dependencies: Dict[LspSymbol, bool]
    blacklist: Set[LspSymbol]
    dependencies: Optional[Dict[LspSymbol, bool]]
    context: Optional[LspSymbol]

    def __init__(self, indexer: LspIndexer) -> None:
        BaseLspSymbolVisitor.__init__(self, indexer)
        self.cyclic_dependencies = OrderedDict()
        self.blacklist = set()
        self.dependencies = None
        self.context = None

    @contextmanager
    def push_context(self, context: LspSymbol) -> Iterator[Dict[LspSymbol, bool]]:
        old_dependencies = self.dependencies
        old_context = self.context

        self.dependencies = OrderedDict()
        self.context = context

        yield self.dependencies

        self.dependencies = old_dependencies
        self.context = old_context

    def extract(self) -> None:
        for symbol in self.indexer.name_index.values():
            self.visit_symbol(symbol)
            if symbol in self.cyclic_dependencies \
               and symbol.dependencies is not None:
                if symbol.cycles is None:
                    symbol.cycles = set()
                for dependency in symbol.dependencies:
                    if dependency.dependencies is not None \
                       and symbol in dependency.dependencies:
                        symbol.cycles.add(dependency)
                        if dependency.cycles is None:
                            dependency.cycles = set()
                        dependency.cycles.add(symbol)

    def visit_symbol(self, symbol: LspSymbol) -> None:
        with self.nested_name_as(symbol.name) as symbol_name:
            if symbol.dependencies is None:
                symbol.dependencies = []  # avoid infinite recursion on cyclic dependencies
                with self.push_context(symbol) as dependencies:
                    match symbol.kind:
                        case LspSymbolKind.ALIAS:
                            BaseLspSpecVisitor.visit_type_alias(self, symbol.spec)
                        case LspSymbolKind.ARRAY:
                            BaseLspSpecVisitor.visit_array(self, symbol.spec)
                        case LspSymbolKind.BASE:
                            BaseLspSpecVisitor.visit_base(self, symbol.spec)
                        case LspSymbolKind.ENUMERATION:
                            BaseLspSpecVisitor.visit_enumeration(self, symbol.spec)
                        case LspSymbolKind.INNER:
                            BaseLspSpecVisitor.visit_inner(self, symbol.spec)
                        case LspSymbolKind.MAP:
                            BaseLspSpecVisitor.visit_map(self, symbol.spec)
                        case LspSymbolKind.NOTIFICATION:
                            BaseLspSpecVisitor.visit_notification(self, symbol.spec)
                        case LspSymbolKind.REFERENCE:
                            BaseLspSpecVisitor.visit_reference(self, symbol.spec)
                        case LspSymbolKind.REQUEST:
                            BaseLspSpecVisitor.visit_request(self, symbol.spec)
                        case LspSymbolKind.STRUCTURE:
                            BaseLspSpecVisitor.visit_structure(self, symbol.spec)
                        case LspSymbolKind.TUPLE:
                            BaseLspSpecVisitor.visit_tuple(self, symbol.spec)
                        case LspSymbolKind.UNION:
                            BaseLspSpecVisitor.visit_union(self, symbol.spec)
                    symbol.dependencies.extend(dependencies.keys())
            BaseLspSymbolVisitor.visit_symbol(self, symbol)

    @dependency_visitor
    def visit_alias_symbol(self, alias_symbol: LspSymbol) -> None:
        BaseLspSymbolVisitor.visit_alias_symbol(self, alias_symbol)

    @dependency_visitor
    def visit_array_symbol(self, array_symbol: LspSymbol) -> None:
        BaseLspSymbolVisitor.visit_array_symbol(self, array_symbol)

    @dependency_visitor
    def visit_base_symbol(self, base_symbol: LspSymbol) -> None:
        BaseLspSymbolVisitor.visit_base_symbol(self, base_symbol)

    @dependency_visitor
    def visit_enumeration_symbol(self, enum_symbol: LspSymbol) -> None:
        BaseLspSymbolVisitor.visit_enumeration_symbol(self, enum_symbol)

    @dependency_visitor
    def visit_inner_symbol(self, inner_symbol: LspSymbol) -> None:
        BaseLspSymbolVisitor.visit_inner_symbol(self, inner_symbol)

    @dependency_visitor
    def visit_map_symbol(self, map_symbol: LspSymbol) -> None:
        BaseLspSymbolVisitor.visit_map_symbol(self, map_symbol)

    @dependency_visitor
    def visit_notification_symbol(self, notification_symbol: LspSymbol) -> None:
        BaseLspSymbolVisitor.visit_notification_symbol(self, notification_symbol)

    @dependency_visitor
    def visit_reference_symbol(self, reference_symbol: LspSymbol) -> None:
        BaseLspSymbolVisitor.visit_reference_symbol(self, reference_symbol)

    @dependency_visitor
    def visit_request_symbol(self, request_symbol: LspSymbol) -> None:
        BaseLspSymbolVisitor.visit_request_symbol(self, request_symbol)

    @dependency_visitor
    def visit_structure_symbol(self, struct_symbol: LspSymbol) -> None:
        BaseLspSymbolVisitor.visit_structure_symbol(self, struct_symbol)

    @dependency_visitor
    def visit_tuple_symbol(self, tuple_symbol: LspSymbol) -> None:
        BaseLspSymbolVisitor.visit_tuple_symbol(self, tuple_symbol)

    @dependency_visitor
    def visit_union_symbol(self, union_symbol: LspSymbol) -> None:
        BaseLspSymbolVisitor.visit_union_symbol(self, union_symbol)


class LspSymbolResolver(LspSpecVisitor[LspSymbol], NestedContext):
    indexer: LspIndexer

    def __init__(self, indexer: LspIndexer) -> None:
        LspSpecVisitor.__init__(self)
        NestedContext.__init__(self)
        self.indexer = indexer

    def resolve_all(self) -> None:
        for symbol in self.indexer.name_index.values():
            self.resolve(symbol)

    def resolve(self, symbol_or_name: Optional[Union[str, LspSymbol]] = None) -> LspSymbol:
        match symbol_or_name:
            case str():
                symbol_name: str = symbol_or_name
                symbol: LspSymbol = self.indexer.name_index[symbol_name]
            case LspSymbol():
                symbol: LspSymbol = symbol_or_name
            case _:
                raise ValueError(f'Unsupported symbol type ({type(symbol_or_name)}): {symbol_or_name}')
        if symbol.resolution is None:
            match symbol.kind:
                case LspSymbolKind.ALIAS:
                    symbol.resolution = self.visit_alias_symbol(symbol)
                case LspSymbolKind.REFERENCE:
                    symbol.resolution = self.visit_reference_symbol(symbol)
                case _:
                    symbol.resolution = symbol
        return symbol.resolution

    def visit_alias_symbol(self, alias_symbol: LspSymbol) -> LspSymbol:
        alias_spec = alias_symbol.spec
        return self.visit_type_alias(alias_spec)

    def visit_reference_symbol(self, reference_symbol: LspSymbol) -> LspSymbol:
        reference_spec = reference_symbol.spec
        return self.visit_type(reference_spec)

    def visit_base(self, base_spec: LspSpec) -> LspSymbol:
        base_name = base_spec["name"]
        return self.resolve(base_name)

    def visit_reference(self, reference_spec: LspSpec) -> LspSymbol:
        type_name = reference_spec["name"]
        return self.resolve(type_name)

    def visit_enumeration(self, enum_spec: LspSpec) -> LspSymbol:
        enum_name = enum_spec["name"]
        return self.resolve(enum_name)

    def visit_array(self, array_spec: LspSpec) -> LspSymbol:
        array_name = self.nested_name()
        return self.resolve(array_name)

    def visit_map(self, map_spec: LspSpec) -> LspSymbol:
        map_name = self.nested_name()
        return self.resolve(map_name)

    def visit_and(self, and_spec: LspSpec) -> LspSymbol:
        and_name = self.nested_name()
        return self.resolve(and_name)

    def visit_union(self, union_spec: LspSpec) -> LspSymbol:
        union_name = self.nested_name()
        return self.resolve(union_name)

    def visit_tuple(self, tuple_spec: LspSpec) -> LspSymbol:
        tuple_name = self.nested_name()
        return self.resolve(tuple_name)

    def visit_inner(self, inner_spec: LspSpec) -> LspSymbol:
        nested_name = self.nested_name()
        return self.resolve(nested_name)

    def visit_structure(self, struct_spec: LspSpec) -> LspSymbol:
        struct_name: str = struct_spec["name"]
        return self.resolve(struct_name)

    def visit_type_alias(self, alias_spec: LspSpec) -> LspSymbol:
        alias_name = alias_spec["name"]
        with self.nested_name_as(alias_name):
            type_spec: LspSpec = alias_spec["type"]
            return self.visit_type(type_spec)

    def visit_request(self, request_spec: LspSpec) -> LspSymbol:
        request_method = request_spec["method"]
        return self.resolve(request_method)

    def visit_notification(self, notification_spec: LspSpec) -> LspSymbol:
        notification_method = notification_spec["method"]
        return self.resolve(notification_method)

    def visit_string_literal(self, string_spec: LspSpec) -> LspSymbol:
        raise NotImplementedError

    def visit_integer_literal(self, integer_spec: LspSpec) -> LspSymbol:
        raise NotImplementedError

    def visit_boolean_literal(self, boolean_spec: LspSpec) -> LspSymbol:
        raise NotImplementedError


class LspStructFieldExpander:
    indexer: LspIndexer
    resolver: LspSymbolResolver

    def __init__(self, indexer: LspIndexer) -> None:
        self.indexer = indexer
        self.resolver = LspSymbolResolver(indexer)

    def expand(self) -> None:
        for struct_symbol in self.indexer.kind_index[LspSymbolKind.STRUCTURE]:
            self.visit_structure_symbol(struct_symbol)
        for inner_symbol in self.indexer.kind_index[LspSymbolKind.INNER]:
            self.visit_inner_symbol(inner_symbol)

    def visit_structure_symbol(self, struct_symbol: LspSymbol) -> None:
        struct_fields = chain()
        pending: Deque[LspSymbol] = deque([struct_symbol])
        while len(pending) > 0:
            pending_symbol = pending.popleft()
            pending_spec = pending_symbol.spec
            struct_fields = chain(
                pending_spec["properties"],
                struct_fields
            )
            mixin_specs = pending_spec.get("mixins", None)
            if mixin_specs is not None:
                for mixin_spec in mixin_specs:
                    mixin_symbol = self.resolver.resolve(mixin_spec["name"])
                    pending.append(mixin_symbol)
        struct_symbol.fields = list(struct_fields)

    def visit_inner_symbol(self, inner_symbol: LspSymbol) -> None:
        inner_spec = inner_symbol.spec
        value_spec: LspSpec = inner_spec["value"]
        inner_symbol.fields = list(value_spec["properties"])


class LspAnalysisPipeline:
    schema: LspSpec
    indexer: LspIndexer
    dependency_extractor: LspDependencyExtractor
    struct_field_expander: LspStructFieldExpander
    symbol_resolver: LspSymbolResolver

    def __init__(self, schema: LspSpec) -> None:
        self.schema = schema
        self.indexer = LspIndexer(schema)
        self.dependency_extractor = LspDependencyExtractor(self.indexer)
        self.struct_field_expander = LspStructFieldExpander(self.indexer)
        self.symbol_resolver = LspSymbolResolver(self.indexer)

    def run(self) -> None:
        self.indexer.index()
        self.dependency_extractor.extract()
        self.struct_field_expander.expand()
        self.symbol_resolver.resolve_all()


class LspTopologicalIterator(Iterable[LspSymbol]):
    dependencies_by_symbol: Dict[LspSymbol, List[LspSymbol]]
    dependents_by_symbol: Dict[LspSymbol, List[LspSymbol]]

    def __init__(self, indexer: LspIndexer) -> None:
        self.dependencies_by_symbol = OrderedDict()
        self.dependents_by_symbol = OrderedDict()

        for symbol in indexer.name_index.values():
            if symbol.dependencies is not None:
                dependencies = []
                for dependency in symbol.dependencies:
                    dependencies.append(dependency)
                    dependents = self.dependents_by_symbol.get(dependency, None)
                    if dependents is None:
                        dependents = []
                        self.dependents_by_symbol[dependency] = dependents
                    dependents.append(symbol)
                self.dependencies_by_symbol[symbol] = dependencies
            else:
                raise ValueError(
                    f'No dependencies were defined for symbol: {symbol.name}'
                )

    def __iter__(self) -> Iterator[LspSymbol]:
        yield from self.sort()

    def resolve(
            self,
            resolved: Set[LspSymbol],
            symbol: LspSymbol
    ) -> Iterator[LspSymbol]:
        if symbol not in resolved:
            yield symbol
            resolved.add(symbol)
            dependents = self.dependents_by_symbol.get(symbol, None)
            if dependents is not None:
                for dependent in self.dependents_by_symbol[symbol]:
                    dependent_dependencies = \
                        self.dependencies_by_symbol.get(dependent, None)
                    if dependent_dependencies is not None:
                        if symbol in dependent_dependencies:
                            dependent_dependencies.remove(symbol)
                        if len(dependent_dependencies) == 0:
                            yield from self.resolve(resolved, dependent)
                del self.dependents_by_symbol[symbol]
            if symbol in self.dependencies_by_symbol:
                del self.dependencies_by_symbol[symbol]

    def sort(self) -> Iterator[LspSymbol]:
        resolved: Set[LspSymbol] = set()
        for symbol, dependencies in list(self.dependencies_by_symbol.items()):
            if len(dependencies) == 0:
                yield from self.resolve(resolved, symbol)
        num_failed = len(self.dependencies_by_symbol)
        if num_failed > 0:
            raise RuntimeError(f'Failed to sort {num_failed} dependencies.')



import argparse
from concurrent.futures import ThreadPoolExecutor
from functools import partial
import hashlib
import itertools
import json
import logging
import os
import re
import pathlib
import pprint
import shutil
import subprocess
import sys
import toml
from typing import Any, Mapping, List, Union

level = logging.DEBUG
log = logging.getLogger(__name__)
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(logging.Formatter('%(message)s'))
handler.setLevel(level)
log.addHandler(handler)
log.setLevel(level)


TESTER_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__)))
LIBASR_DIR = os.path.dirname(TESTER_DIR)
SRC_DIR = os.path.dirname(LIBASR_DIR)
if "lpython" in SRC_DIR:
    SRC_DIR = os.path.join(os.path.dirname(os.path.dirname(SRC_DIR)), "src")
ROOT_DIR = os.path.dirname(SRC_DIR)

no_color = False

class RunException(Exception):
    pass


class ExecuteException(Exception):
    pass


class style:
    reset = 0
    bold = 1
    dim = 2
    italic = 3
    underline = 4
    blink = 5
    rblink = 6
    reversed = 7
    conceal = 8
    crossed = 9


class fg:
    black = 30
    red = 31
    green = 32
    yellow = 33
    blue = 34
    magenta = 35
    cyan = 36
    gray = 37
    reset = 39


def color(value):
    return "\033[" + str(int(value)) + "m"


def check():
    return f"{(color(fg.green)+color(style.bold))}✓ {color(fg.reset)+color(style.reset)}"


def bname(base, cmd, filename):
    hstring = cmd
    if filename:
        hstring += filename
    h = hashlib.sha224(hstring.encode()).hexdigest()[:7]
    if filename:
        bname = os.path.basename(filename)
        bname, _ = os.path.splitext(bname)
        return f"{base}-{bname}-{h}"
    else:
        return f"{base}-{h}"


def _compare_eq_dict(
    left: Mapping[Any, Any], right: Mapping[Any, Any], verbose: int = 0
) -> List[str]:
    explanation: List[str] = []
    set_left = set(left)
    set_right = set(right)
    common = set_left.intersection(set_right)
    same = {k: left[k] for k in common if left[k] == right[k]}
    if same and verbose < 2:
        explanation += ["Omitting %s identical items" % len(same)]
    elif same:
        explanation += ["Common items:"]
        explanation += pprint.pformat(same).splitlines()
    diff = {k for k in common if left[k] != right[k]}
    if diff:
        explanation += ["ACTUAL != REFERENCE"]
        explanation += ["Differing items:"]
        for k in diff:
            explanation += [repr({k: left[k]}) + " != " + repr({k: right[k]})]
    extra_left = set_left - set_right
    len_extra_left = len(extra_left)
    if len_extra_left:
        explanation.append(
            "Left contains %d more item%s:"
            % (len_extra_left, "" if len_extra_left == 1 else "s")
        )
        explanation.extend(
            pprint.pformat({k: left[k] for k in extra_left}).splitlines()
        )
    extra_right = set_right - set_left
    len_extra_right = len(extra_right)
    if len_extra_right:
        explanation.append(
            "Right contains %d more item%s:"
            % (len_extra_right, "" if len_extra_right == 1 else "s")
        )
        explanation.extend(
            pprint.pformat({k: right[k] for k in extra_right}).splitlines()
        )
    return explanation

def test_for_duplicates(test_data):
    tests = test_data["test"]
    filenames = [t["filename"] for t in tests]
    if len(set(filenames)) != len(filenames):
        print("There are duplicate test filenames:")
        duplicates = [item for item in set(filenames) if filenames.count(item) > 1]
        print(duplicates)
        sys.exit(1)

def fixdir(s: bytes) -> bytes:
    local_dir = os.getcwd()
    return s.replace(local_dir.encode(), "$DIR".encode())

def fix_datalayout(s: bytes) -> bytes:
    return re.sub(
        rb'target datalayout = "[^"]*"',
        b'target datalayout = "..."',
        s
    )


def unl_loop_del(b):
    return b.replace(bytes('\r\n', encoding='utf-8'),
                     bytes('\n', encoding='utf-8'))


def run(basename: str, cmd: Union[pathlib.Path, str],
        out_dir: Union[pathlib.Path, str], infile=None, extra_args=None):
    """
    Runs the `cmd` and collects stdout, stderr, exit code.

    The stdout, stderr and outfile are saved in the `out_dir` directory and
    all metadata is saved in a json file, whose path is returned from the
    function.

    The idea is to use this function to test the compiler by running it with
    an option to save the AST, ASR or LLVM IR or binary, and then ensure that
    the output does not change.

    Arguments:

    basename ... name of the run
    cmd ........ command to run, can use {infile} and {outfile}
    out_dir .... output directory to store output
    infile ..... optional input file. If present, it will check that it exists
                 and hash it.
    extra_args . extra arguments, not part of the hash

    Examples:

    >>> run("cat2", "cat tests/cat.txt > {outfile}", "output", "tests/cat.txt")
    >>> run("ls4", "ls --wrong-option", "output")

    """
    assert basename is not None and basename != ""
    pathlib.Path(out_dir).mkdir(parents=True, exist_ok=True)
    if infile and not os.path.exists(infile):
        raise RunException("The input file %s does not exist" % (infile))
    outfile = os.path.join(out_dir, basename + "." + "out")

    infile = infile.replace("\\\\", "\\").replace("\\", "/")

    cmd2 = cmd.format(infile=infile, outfile=outfile)
    if extra_args:
        cmd2 += " " + extra_args
    r = subprocess.run(cmd2, shell=True,
                       stdout=subprocess.PIPE,
                       stderr=subprocess.PIPE)
    if not os.path.exists(outfile):
        outfile = None
    if len(r.stdout):
        stdout_file = os.path.join(out_dir, basename + "." + "stdout")
        open(stdout_file, "wb").write(fix_datalayout(fixdir(r.stdout)))
    else:
        stdout_file = None
    if len(r.stderr):
        stderr_file = os.path.join(out_dir, basename + "." + "stderr")
        open(stderr_file, "wb").write(fix_datalayout(fixdir(r.stderr)))
    else:
        stderr_file = None

    if infile:
        temp = unl_loop_del(open(infile, "rb").read())
        infile_hash = hashlib.sha224(temp).hexdigest()
    else:
        infile_hash = None
    if outfile:
        temp = unl_loop_del(open(outfile, "rb").read())
        outfile_hash = hashlib.sha224(temp).hexdigest()
        outfile = os.path.basename(outfile)
    else:
        outfile_hash = None
    if stdout_file:
        temp = unl_loop_del(open(stdout_file, "rb").read())
        stdout_hash = hashlib.sha224(temp).hexdigest()
        stdout_file = os.path.basename(stdout_file)
    else:
        stdout_hash = None
    if stderr_file:
        temp = unl_loop_del(open(stderr_file, "rb").read())
        stderr_hash = hashlib.sha224(temp).hexdigest()
        stderr_file = os.path.basename(stderr_file)
    else:
        stderr_hash = None
    data = {
        "basename": basename,
        "cmd": cmd,
        "infile": infile,
        "infile_hash": infile_hash,
        "outfile": outfile,
        "outfile_hash": outfile_hash,
        "stdout": stdout_file,
        "stdout_hash": stdout_hash,
        "stderr": stderr_file,
        "stderr_hash": stderr_hash,
        "returncode": r.returncode,
    }
    json_file = os.path.join(out_dir, basename + "." + "json")
    json.dump(data, open(json_file, "w"), indent=4)
    return json_file


def get_error_diff(reference_file, output_file, full_err_str, field="") -> str:
    import tempfile

    ref_exists = os.path.exists(reference_file)
    out_exists = os.path.exists(output_file)

    # For stderr, always show as diff (even if one file is missing)
    if field == "stderr":
        # Create temp files for missing sides
        temp_ref = None
        temp_out = None
        ref_to_use = reference_file
        out_to_use = output_file

        if not ref_exists:
            temp_ref = tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.stderr')
            temp_ref.close()
            ref_to_use = temp_ref.name

        if not out_exists:
            temp_out = tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.stderr')
            temp_out.close()
            out_to_use = temp_out.name

        # Run diff
        diff_list = subprocess.Popen(
            f"diff {ref_to_use} {out_to_use}",
            stdout=subprocess.PIPE,
            shell=True,
            encoding='utf-8')
        diff_str = ""
        diffs = diff_list.stdout.readlines()
        for d in diffs:
            diff_str += d

        full_err_str += f"\n=== STDERR DIFF ===\n"
        if not ref_exists:
            full_err_str += f"Reference: (missing - expected no stderr)\n"
        else:
            full_err_str += f"Reference: {reference_file}\n"
        if not out_exists:
            full_err_str += f"Output:    (missing - no stderr produced)\n"
        else:
            full_err_str += f"Output:    {output_file}\n"
        full_err_str += diff_str if diff_str else "(files are identical)\n"

        # Cleanup temp files
        if temp_ref:
            os.unlink(temp_ref.name)
        if temp_out:
            os.unlink(temp_out.name)

        return full_err_str

    # For stdout and outfile, use the original behavior with content preview
    if not ref_exists and not out_exists:
        full_err_str += f"\nNeither reference nor output file exists\n"
        full_err_str += f"Reference: {reference_file}\n"
        full_err_str += f"Output:    {output_file}\n"
        return full_err_str

    if not out_exists:
        full_err_str += f"\n=== OUTPUT FILE MISSING ===\n"
        full_err_str += f"Expected output file: {output_file}\n"
        full_err_str += f"\n=== EXPECTED CONTENT (from reference) ===\n"
        try:
            with open(reference_file, 'r', encoding='utf-8', errors='replace') as f:
                content = f.read()
                full_err_str += content if content else "(empty file)\n"
        except Exception as e:
            full_err_str += f"(could not read reference file: {e})\n"
        return full_err_str

    if not ref_exists:
        full_err_str += f"\n=== REFERENCE FILE MISSING ===\n"
        full_err_str += f"Expected reference file: {reference_file}\n"
        full_err_str += f"\n=== ACTUAL OUTPUT ===\n"
        try:
            with open(output_file, 'r', encoding='utf-8', errors='replace') as f:
                content = f.read()
                full_err_str += content if content else "(empty file)\n"
        except Exception as e:
            full_err_str += f"(could not read output file: {e})\n"
        return full_err_str

    # Both files exist - show diff
    diff_list = subprocess.Popen(
        f"diff {reference_file} {output_file}",
        stdout=subprocess.PIPE,
        shell=True,
        encoding='utf-8')
    diff_str = ""
    diffs = diff_list.stdout.readlines()
    for d in diffs:
        diff_str += d

    full_err_str += f"\n=== DIFF ===\n"
    full_err_str += f"Reference: {reference_file}\n"
    full_err_str += f"Output:    {output_file}\n"
    full_err_str += diff_str if diff_str else "(files are identical)\n"

    return full_err_str


def do_update_reference(jo, jr, do):
    os.makedirs(os.path.dirname(jr), exist_ok=True)
    shutil.copyfile(jo, jr)
    for f in ["outfile", "stdout", "stderr"]:
        if do[f]:
            f_o = os.path.join(os.path.dirname(jo), do[f])
            f_r = os.path.join(os.path.dirname(jr), do[f])
            shutil.copyfile(f_o, f_r)

def do_verify_reference_hash(jr, dr, s):
    for f in ["outfile", "stdout", "stderr"]:
        if dr[f]:
            f_r = os.path.join(os.path.dirname(jr), dr[f])
            temp = unl_loop_del(open(f_r, "rb").read())
            f_r_hash = hashlib.sha224(temp).hexdigest()
            if (f_r_hash != dr[f + "_hash"]):
                # This string builds up the error message.
                # Print test name in red in the beginning.
                # More information is added afterwards.
                full_err_str = f"\n{(color(fg.red)+color(style.bold))}{s}{color(fg.reset)+color(style.reset)}\n"
                full_err_str += "The generated hash for the reference file and its committed hash are different\n"
                full_err_str += "Reference File: " + f_r + "\n"
                full_err_str += "Reference Json File: " + jr + "\n"
                full_err_str += "Reference File Hash Expected: " + f_r_hash + "\n"
                full_err_str += "Reference File Hash Found: " + dr[f + "_hash"] + "\n"
                raise RunException("Verifying reference hash failed." +
                    full_err_str)

def run_test(testname, basename, cmd, infile, update_reference=False,
            verify_hash=False, extra_args=None):
    """
    Runs the test `cmd` and compare against reference results.

    The `cmd` is executed via `run` (passing in `basename` and `infile`) and
    the output is saved in the `output` directory. The generated json file is
    then compared against reference results and if it differs, the
    RunException is thrown.

    Arguments:

    basename ........... name of the run
    cmd ................ command to run, can use {infile} and {outfile}
    infile ............. optional input file. If present, it will check that
                         it exists and hash it.
    update_reference ... if True, it will copy the output into the reference
                         directory as reference results, overwriting old ones
    verify_hash ...... if True, it will check the hash in the committed
                         json file and the hash for the committed references
                         directory as reference results, overwriting old ones
    extra_args ......... Extra arguments to append to the command that are not
                         part of the hash

    Examples:

    >>> run_test("cat12", "cat {infile} > {outfile}", "cat.txt",
    ...     update_reference=True)
    >>> run_test("cat12", "cat {infile} > {outfile}", "cat.txt")
    """
    s = f"{testname} * {basename}"
    basename = bname(basename, cmd, infile)
    infile = os.path.join("tests", infile)
    jo = run(basename, cmd, os.path.join("tests", "output"), infile=infile,
             extra_args=extra_args)
    jr = os.path.join("tests", "reference", os.path.basename(jo))
    if not os.path.exists(jo):
        raise FileNotFoundError(
            f"The output json file '{jo}' for {testname} does not exist")

    try:
        do = json.load(open(jo))
    except json.decoder.JSONDecodeError:
        print("JSON failed to be decoded")
        print(f"Filename: {jo}")
        raise
    if update_reference:
        do_update_reference(jo, jr, do)
        return

    if not os.path.exists(jr):
        raise FileNotFoundError(
            f"The reference json file '{jr}' for {testname} does not exist")

    dr = json.load(open(jr))

    if verify_hash:
        do_verify_reference_hash(jr, dr, s)
        return

    if do != dr:
        # This string builds up the error message. Print test name in red in the beginning.
        # More information is added afterwards.
        full_err_str = f"\n{(color(fg.red)+color(style.bold))}{s}{color(fg.reset)+color(style.reset)}\n"
        e = _compare_eq_dict(do, dr)
        full_err_str += "The JSON metadata differs against reference results\n"
        full_err_str += "Reference JSON: " + jr + "\n"
        full_err_str += "Output JSON:    " + jo + "\n"
        full_err_str += "\n".join(e)

        for field in ["outfile", "stdout", "stderr"]:
            hash_field = field + "_hash"
            if not do[hash_field] and dr[hash_field]:
                full_err_str += f"\n=== MISSING OUTPUT {field.upper()} ===\n"
                full_err_str += f"Expected {field} to be generated but it was not.\n"
                reference_file = os.path.join("tests", "reference", dr[field])
                if os.path.exists(reference_file):
                    if field == "stdout":
                        # For stdout, show only first 10 lines
                        full_err_str += f"\n=== EXPECTED CONTENT (first 10 lines from reference) ===\n"
                        try:
                            with open(reference_file, 'r', encoding='utf-8', errors='replace') as f:
                                lines = f.readlines()
                                preview = ''.join(lines[:10])
                                full_err_str += preview if preview else "(empty file)\n"
                                if len(lines) > 10:
                                    full_err_str += f"\n... ({len(lines) - 10} more lines, see {reference_file})\n"
                        except Exception as e:
                            full_err_str += f"(could not read reference file: {e})\n"
                    else:
                        # For stderr and outfile, use get_error_diff which will handle diff
                        output_file = os.path.join("tests", "output", do[field] if do[field] else "missing")
                        full_err_str = get_error_diff(
                            reference_file, output_file, full_err_str, field)
            elif not dr[hash_field] and do[hash_field]:
                full_err_str += f"\n=== UNEXPECTED OUTPUT {field.upper()} ===\n"
                full_err_str += f"Did not expect {field} to be generated but it was.\n"
                output_file = os.path.join("tests", "output", do[field])
                if os.path.exists(output_file):
                    if field == "stdout":
                        # For stdout, show only first 10 lines
                        full_err_str += f"\n=== ACTUAL OUTPUT (first 10 lines) ===\n"
                        try:
                            with open(output_file, 'r', encoding='utf-8', errors='replace') as f:
                                lines = f.readlines()
                                preview = ''.join(lines[:10])
                                full_err_str += preview if preview else "(empty file)\n"
                                if len(lines) > 10:
                                    full_err_str += f"\n... ({len(lines) - 10} more lines, see {output_file})\n"
                        except Exception as e:
                            full_err_str += f"(could not read output file: {e})\n"
                    else:
                        # For stderr and outfile, use get_error_diff
                        reference_file = os.path.join("tests", "reference", dr[field] if dr[field] else "missing")
                        full_err_str = get_error_diff(
                            reference_file, output_file, full_err_str, field)
            elif do[hash_field] != dr[hash_field]:
                output_file = os.path.join("tests", "output", do[field])
                reference_file = os.path.join("tests", "reference", dr[field])
                full_err_str = get_error_diff(
                    reference_file, output_file, full_err_str, field)
        raise RunException(
            "Testing with reference output failed." +
            full_err_str)
    if no_color:
        log.debug(s + " PASS")
    else:
        log.debug(s + " " + check())


def tester_main(compiler, single_test, is_lcompilers_executable_installed=False):
    parser = argparse.ArgumentParser(description=f"{compiler} Test Suite")
    parser.add_argument("-u", "--update", action="store_true",
                        help="update all reference results")
    parser.add_argument("-vh", "--verify-hash", action="store_true",
                        help="Verify all reference hashes")
    parser.add_argument("-l", "--list", action="store_true",
                        help="list all tests")
    parser.add_argument("-t", "--test",
                        action="append", nargs="*",
                        help="Run specific tests")
    parser.add_argument("-b", "--backend",
                        action="append", nargs="*",
                        help="Run specific backends")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="increase test verbosity")
    parser.add_argument("--exclude-test", metavar="TEST",
                        action="append", nargs="*",
                        help="Exclude specific tests"),
    parser.add_argument("--exclude-backend", metavar="BACKEND",
                        action="append", nargs="*",
                        help="Exclude specific backends, only works when -b is not specified"),
    parser.add_argument("--no-llvm", action="store_true",
                        help="Skip LLVM tests")
    parser.add_argument("--skip-run-with-dbg", action="store_true",
                        help="Skip runtime tests with debugging information enabled")
    parser.add_argument("-s", "--sequential", action="store_true",
                        help="Run all tests sequentially")
    parser.add_argument("--no-color", action="store_true",
                    help="Turn off colored tests output")
    args = parser.parse_args()
    update_reference = args.update
    verify_hash = args.verify_hash
    list_tests = args.list
    specific_tests = list(
        itertools.chain.from_iterable(
            args.test)) if args.test else None
    specific_backends = set(
        itertools.chain.from_iterable(
            args.backend)) if args.backend else None
    excluded_tests = list(itertools.chain.from_iterable(
        args.exclude_test)) if args.exclude_test else None
    excluded_backends = set(itertools.chain.from_iterable(
        args.exclude_backend)) if args.exclude_backend and specific_backends is None else None
    verbose = args.verbose
    no_llvm = args.no_llvm
    skip_run_with_dbg = args.skip_run_with_dbg
    global no_color
    no_color = args.no_color

    # While updating references, only wipe the whole reference directory if the
    # user is updating the entire suite (no -t filters). For a targeted update
    # (via -t), deleting everything is unexpected and breaks unrelated tests.
    if update_reference and not specific_tests:
        log.debug("REMOVE: old test references")
        cmd = "rm -rf ./tests/reference/*"
        log.debug(f"+ {cmd}")
        process = subprocess.run(cmd, shell=True)
        if process.returncode != 0:
            print("Removing Old test references failed!")
            exit(1)

    # So that the tests find the `lcompiler` executable
    if not is_lcompilers_executable_installed:
        os.environ["PATH"] = os.path.join(SRC_DIR, "bin") \
            + os.pathsep + os.environ["PATH"]
    test_data = toml.load(open(os.path.join(ROOT_DIR, "tests", "tests.toml")))
    test_for_duplicates(test_data)
    filtered_tests = test_data["test"]
    if specific_tests:
        filtered_tests = [test for test in filtered_tests if any(
            re.search(t, test["filename"]) for t in specific_tests)]
    if excluded_tests:
        filtered_tests = [test for test in filtered_tests if not any(
            re.search(t, test["filename"]) for t in excluded_tests)]
    if specific_backends:
        filtered_tests = [
            test for test in filtered_tests if any(
                b in test for b in specific_backends)]
    if excluded_backends:
        filtered_tests = [test for test in filtered_tests if any(
            b not in excluded_backends and b != "filename" for b in test)]

    for test in filtered_tests:
        if 'extrafiles' in test:
            single_test(test,
                update_reference=update_reference,
                verify_hash=verify_hash,
                specific_backends=specific_backends,
                excluded_backends=excluded_backends,
                verbose=verbose,
                no_llvm=no_llvm,
                skip_run_with_dbg=True,
                no_color=True)
    filtered_tests = [test for test in filtered_tests if 'extrafiles' not in test]

    if args.sequential:
        for test in filtered_tests:
            single_test(test,
                        update_reference=update_reference,
                        verify_hash=verify_hash,
                        specific_backends=specific_backends,
                        excluded_backends=excluded_backends,
                        verbose=verbose,
                        no_llvm=no_llvm,
                        skip_run_with_dbg=skip_run_with_dbg,
                        no_color=no_color)
    # run in parallel
    else:
        single_tester_partial_args = partial(
            single_test,
            update_reference=update_reference,
            verify_hash=verify_hash,
            specific_backends=specific_backends,
            excluded_backends=excluded_backends,
            verbose=verbose,
            no_llvm=no_llvm,
            skip_run_with_dbg=skip_run_with_dbg,
            no_color=no_color)
        with ThreadPoolExecutor() as ex:
            futures = ex.map(single_tester_partial_args, filtered_tests)
            for f in futures:
                if not f:
                    ex.shutdown(wait=False)
    if list_tests:
        return

    if update_reference:
        log.info("Test references updated.")
    elif verify_hash:
        if no_color:
            log.info("Test references hash verfied.")
        else:
            log.info(
                f"{(color(fg.green) + color(style.bold))}Test references hash verfied."
                f"{color(fg.reset) + color(style.reset)}")
    else:
        if no_color:
            log.info("TESTS PASSED")
        else:
            log.info(
                f"{(color(fg.green) + color(style.bold))}TESTS PASSED"
                f"{color(fg.reset) + color(style.reset)}")



import argparse
from typing import Any, Dict, Tuple

from llanguage_server.cxx.lsp_language_server_header_generator import \
    CPlusPlusLspLanguageServerHeaderGenerator
from llanguage_server.cxx.lsp_language_server_source_generator import \
    CPlusPlusLspLanguageServerSourceGenerator
from llanguage_server.cxx.lsp_specification_header_generator import \
    CPlusPlusSpecificationHeaderGenerator
from llanguage_server.cxx.lsp_specification_source_generator import \
    CPlusPlusSpecificationSourceGenerator
from llanguage_server.cxx.lsp_transformer_header_generator import \
    CPlusPlusLspTransformerHeaderGenerator
from llanguage_server.cxx.lsp_transformer_source_generator import \
    CPlusPlusLspTransformerSourceGenerator
from llanguage_server.lsp_code_generator import LspCodeGenerator


class CPlusPlusLspCodeGenerator(LspCodeGenerator):
    namespace: str = "LCompilers::LanguageServerProtocol"

    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)

    def generate_specification_header(self) -> None:
        code_generator = CPlusPlusSpecificationHeaderGenerator(
            self.args.output_dir,
            self.schema,
            self.pipeline,
            self.namespace
        )
        with code_generator.open():
            code_generator.generate_code()

    def generate_specification_source(self) -> None:
        code_generator = CPlusPlusSpecificationSourceGenerator(
            self.args.output_dir,
            self.schema,
            self.pipeline,
            self.namespace
        )
        with code_generator.open():
            code_generator.generate_code()

    def generate_specification(self) -> None:
        self.generate_specification_header()
        self.generate_specification_source()

    def generate_transformer_header(self) -> None:
        code_generator = CPlusPlusLspTransformerHeaderGenerator(
            self.args.output_dir,
            self.schema,
            self.pipeline,
            self.namespace
        )
        with code_generator.open():
            code_generator.generate_code()

    def generate_transformer_source(self) -> None:
        code_generator = CPlusPlusLspTransformerSourceGenerator(
            self.args.output_dir,
            self.schema,
            self.pipeline,
            self.namespace
        )
        with code_generator.open():
            code_generator.generate_code()

    def generate_server_header(self) -> None:
        code_generator = CPlusPlusLspLanguageServerHeaderGenerator(
            self.args.output_dir,
            self.schema,
            self.pipeline,
            self.namespace
        )
        with code_generator.open():
            code_generator.generate_code()

    def generate_server_source(self) -> None:
        code_generator = CPlusPlusLspLanguageServerSourceGenerator(
            self.args.output_dir,
            self.schema,
            self.pipeline,
            self.namespace
        )
        with code_generator.open():
            code_generator.generate_code()

    def generate_transformer(self) -> None:
        self.generate_transformer_header()
        self.generate_transformer_source()

    def generate_server(self) -> None:
        self.generate_server_header()
        self.generate_server_source()



from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Union

from lsprotocol.types import (
    CompletionClientCapabilities,
    CompletionClientCapabilitiesCompletionItemType,
    CompletionClientCapabilitiesCompletionItemTypeInsertTextModeSupportType,
    CompletionClientCapabilitiesCompletionItemTypeResolveSupportType,
    CompletionClientCapabilitiesCompletionItemTypeTagSupportType,
    CompletionItemTag, DefinitionClientCapabilities, DefinitionParams,
    DidChangeTextDocumentParams, DocumentFormattingClientCapabilities,
    DocumentHighlightClientCapabilities,
    DocumentRangeFormattingClientCapabilities,
    DocumentSymbolClientCapabilities, HoverClientCapabilities,
    InitializeParams, InsertTextMode, Location, LocationLink, MarkupKind,
    Position, RenameClientCapabilities, RenameParams,
    SemanticTokensClientCapabilities,
    SemanticTokensClientCapabilitiesRequestsType,
    TextDocumentCompletionResponse, TextDocumentContentChangeEvent,
    TextDocumentContentChangeEvent_Type1, TextDocumentContentChangeEvent_Type2,
    TextDocumentDefinitionRequest, TextDocumentDefinitionResponse,
    TextDocumentDocumentHighlightResponse, TextDocumentDocumentSymbolResponse,
    TextDocumentFormattingResponse, TextDocumentHoverResponse,
    TextDocumentIdentifier, TextDocumentPublishDiagnosticsNotification,
    TextDocumentRangeFormattingResponse, TextDocumentRenameRequest,
    TextDocumentRenameResponse, TextDocumentSemanticTokensFullResponse,
    TextDocumentSyncKind, TokenFormat, VersionedTextDocumentIdentifier,
    WorkspaceEdit)

from llanguage_test_client.json_rpc import JsonArray, JsonObject
from llanguage_test_client.lsp_test_client import LspTestClient
from llanguage_test_client.lsp_text_document import LspTextDocument


class LFortranLspTestClient(LspTestClient):

    def __init__(
            self,
            server_path: Path,
            server_params: List[str],
            workspace_path: Optional[Path],
            timeout_ms: float,
            config: Dict[str, Any],
            client_log_path: str,
            stdout_log_path: str,
            stdin_log_path: str
    ) -> None:
        super().__init__(
            server_path,
            server_params,
            workspace_path,
            timeout_ms,
            config,
            client_log_path,
            stdout_log_path,
            stdin_log_path
        )

    def await_validation(self, uri: str, version: int) -> Any:
        def is_validation(message: Any) -> bool:
            if message.get("method", None) == "textDocument/publishDiagnostics":
                params = message["params"]
                return params["uri"] == uri \
                    and params.get("version", None) == version
            return False
        event, _ = self.find_incoming_event(lambda event: is_validation(event.data))
        if event is not None:
            return event.data
        while not self.stop.is_set():
            message = self.receive_message()
            if is_validation(message):
                return message
        return None

    def dispatch_method(self, message: JsonObject) -> None:
        match message["method"]:
            case "textDocument/publishDiagnostics":
                notification = self.converter.structure(
                    message,
                    TextDocumentPublishDiagnosticsNotification
                )
                self.receive_text_document_publish_diagnostics(notification)
                self.respond_to_notification()
            case _:
                super().dispatch_method(message)

    def initialize_params(self) -> InitializeParams:
        params = super().initialize_params()
        text_document = params.capabilities.text_document
        if text_document is not None:
            text_document.completion = CompletionClientCapabilities(
                dynamic_registration=True,
                completion_item=CompletionClientCapabilitiesCompletionItemType(
                    snippet_support=False,
                    commit_characters_support=False,
                    documentation_format=[
                        MarkupKind.PlainText,
                        MarkupKind.Markdown,
                    ],
                    deprecated_support=True,
                    preselect_support=False,
                    tag_support=CompletionClientCapabilitiesCompletionItemTypeTagSupportType(
                        value_set=[
                            CompletionItemTag.Deprecated,
                        ],
                    ),
                    insert_replace_support=False,
                    resolve_support=CompletionClientCapabilitiesCompletionItemTypeResolveSupportType(
                        properties=[],
                    ),
                    insert_text_mode_support=CompletionClientCapabilitiesCompletionItemTypeInsertTextModeSupportType(
                        value_set=[
                            InsertTextMode.AsIs,
                            InsertTextMode.AdjustIndentation,
                        ],
                    ),
                    label_details_support=True,
                ),
                context_support=True,
            )
            text_document.hover = HoverClientCapabilities(
                content_format=[
                    MarkupKind.PlainText,
                    MarkupKind.Markdown,
                ],
            )
            text_document.definition = DefinitionClientCapabilities()
            text_document.rename = RenameClientCapabilities()
            text_document.document_highlight = DocumentHighlightClientCapabilities()
            text_document.document_symbol = DocumentSymbolClientCapabilities(
                hierarchical_document_symbol_support=True,
            )
            text_document.semantic_tokens = SemanticTokensClientCapabilities(
                requests=SemanticTokensClientCapabilitiesRequestsType(
                    full=True,
                ),
                token_types=[
                    "namespace",
                    "type",
                    "class",
                    "enum",
                    "interface",
                    "struct",
                    "typeParameter",
                    "parameter",
                    "variable",
                    "property",
                    "enumMember",
                    "event",
                    "function",
                    "method",
                    "macro",
                    "keyword",
                    "modifier",
                    "comment",
                    "string",
                    "number",
                    "regexp",
                    "operator",
                    "decorator",
                ],
                token_modifiers=[
                    "declaration",
                    "definition",
                    "readonly",
                    "static",
                    "deprecated",
                    "abstract",
                    "async",
                    "modification",
                    "documentation",
                    "defaultLibrary",
                ],
                formats=[TokenFormat.Relative],
            )
            text_document.formatting = DocumentFormattingClientCapabilities()
            text_document.range_formatting = DocumentRangeFormattingClientCapabilities()
        return params

    def receive_text_document_publish_diagnostics(
            self,
            notification: TextDocumentPublishDiagnosticsNotification
    ) -> None:
        pass

    def server_supports_text_document_definition(self) -> bool:
        # NOTE: This returns True if `definition_provider` is True or an instance
        # of `DefinitionOptions`
        return bool(self.server_capabilities.definition_provider)

    def send_text_document_definition(self, params: DefinitionParams) -> int:
        request_id = self.next_request_id()
        request = TextDocumentDefinitionRequest(request_id, params)
        self.send_request(request_id, request,
                          self.receive_text_document_definition)
        return request_id

    def receive_text_document_definition(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        response = self.converter.structure(
            message,
            TextDocumentDefinitionResponse
        )
        loc: Optional[Union[Location, LocationLink]] = None
        match response.result:
            case Location():
                loc = response.result
            case _ if isinstance(response.result, Sequence) \
                 and not isinstance(response.result, str) \
                 and len(response.result) > 0:
                loc = response.result[0]
        doc: Optional[LspTextDocument] = None
        pos: Optional[Position] = None
        match loc:
            case Location():
                doc = self.documents_by_uri.get(loc.uri, None)
                pos = loc.range.start
            case LocationLink():
                doc = self.documents_by_uri.get(loc.target_uri, None)
                pos = loc.target_range.start
        if doc is not None and pos is not None:
            self.active_document = doc
            doc.cursor = (pos.line + 1, pos.character + 1)

    def goto_definition(self, uri: str, line: int, column: int) -> None:
        if self.server_supports_text_document_definition():
            params = DefinitionParams(
                text_document=TextDocumentIdentifier(
                    uri=uri,
                ),
                position=Position(
                    line=line,
                    character=column
                ),
            )
            request_id = self.send_text_document_definition(params)
            self.await_response(request_id)

    def build_custom_request(
            self,
            method: str,
            request_id: int,
            params: Optional[Union[JsonObject, JsonArray]] = None
    ) -> JsonObject:
        request = {
            "jsonrpc": "2.0",
            "method": method,
            "id": request_id,
        }
        if params is not None:
            request["params"] = params
        return request

    def send_document(self, params: JsonObject) -> int:
        request_id = self.next_request_id()
        request = self.build_custom_request("$/document", request_id, params)
        self.send_request(request_id, request, self.receive_document)
        return request_id

    def receive_document(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        pass

    def get_remote_document(self, uri: str) -> JsonObject:
        request_id = self.send_document({
            "uri": uri,
        })
        response = self.await_response(request_id)
        return response["result"]

    def server_supports_text_document_rename(self) -> bool:
        return bool(self.server_capabilities.rename_provider)

    def send_text_document_rename(self, params: RenameParams) -> int:
        request_id = self.next_request_id()
        request = TextDocumentRenameRequest(request_id, params)
        self.send_request(request_id, request, self.receive_text_document_rename)
        return request_id

    def apply(self, workspace_edit: WorkspaceEdit) -> None:
        changes = workspace_edit.changes
        if changes is not None:
            for uri, text_edits in changes.items():
                document = self.get_document("fortran", uri)
                document.apply(text_edits)
                if self.server_supports_text_document_did_change():
                    version = document.bump_version()
                    content_changes: List[TextDocumentContentChangeEvent]
                    if self.server_supports_text_document_sync_kind(
                            TextDocumentSyncKind.Incremental
                    ):
                        content_changes = [
                            TextDocumentContentChangeEvent_Type1(
                                range=text_edit.range,
                                text=text_edit.new_text
                            )
                            for text_edit in text_edits
                        ]
                    else:
                        content_changes = [
                            TextDocumentContentChangeEvent_Type2(
                                text=document.text
                            )
                        ]
                    params = DidChangeTextDocumentParams(
                        text_document=VersionedTextDocumentIdentifier(
                            version=version,
                            uri=uri
                        ),
                        content_changes=content_changes
                    )
                    self.send_text_document_did_change(params)
                document.signal_change()

    def receive_text_document_rename(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        response = self.converter.structure(
            message,
            TextDocumentRenameResponse
        )
        if response.result is not None:
            self.apply(response.result)

    def rename(self, uri: str, line: int, column: int, new_name: str) -> None:
        if self.server_supports_text_document_rename():
            params = RenameParams(
                text_document=TextDocumentIdentifier(
                    uri=uri,
                ),
                position=Position(
                    line=line,
                    character=column,
                ),
                new_name=new_name,
            )
            request_id = self.send_text_document_rename(params)
            self.await_response(request_id)

    def receive_text_document_document_highlight(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        response = self.converter.structure(
            message,
            TextDocumentDocumentHighlightResponse
        )
        if response.result is not None:
            uri = request.params.text_document.uri
            doc = self.get_document("fortran", uri)
            doc.symbol_highlights = response.result

    def receive_text_document_hover(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        response = self.converter.structure(
            message,
            TextDocumentHoverResponse
        )
        if response.result is not None:
            uri = request.params.text_document.uri
            doc = self.get_document("fortran", uri)
            doc.preview = response.result

    def receive_text_document_document_symbol(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        response = self.converter.structure(
            message,
            TextDocumentDocumentSymbolResponse
        )
        uri = request.params.text_document.uri
        doc = self.get_document("fortran", uri)
        doc.symbols = response.result

    def receive_text_document_semantic_tokens_full(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        response = self.converter.structure(
            message,
            TextDocumentSemanticTokensFullResponse
        )
        uri = request.params.text_document.uri
        doc = self.get_document("fortran", uri)
        doc.semantic_highlights = response.result

    def receive_text_document_completion(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        response = self.converter.structure(
            message,
            TextDocumentCompletionResponse
        )
        uri = request.params.text_document.uri
        doc = self.get_document("fortran", uri)
        doc.completions = response.result

    def receive_text_document_formatting(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        response = self.converter.structure(
            message,
            TextDocumentFormattingResponse
        )
        if response.result is not None:
            uri = request.params.text_document.uri
            doc = self.get_document("fortran", uri)
            doc.apply(response.result)

    def receive_text_document_range_formatting(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        response = self.converter.structure(
            message,
            TextDocumentRangeFormattingResponse
        )
        if response.result is not None:
            uri = request.params.text_document.uri
            doc = self.get_document("fortran", uri)
            doc.apply(response.result)

    def send_telemetry(self) -> int:
        request_id = self.next_request_id()
        request = self.build_custom_request("$/telemetry", request_id)
        self.send_request(request_id, request, self.receive_telemetry)
        return request_id

    def receive_telemetry(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        pass

    def get_telemetry(self) -> Any:
        request_id = self.send_telemetry()
        response = self.await_response(request_id)
        return response["result"]



import os
from tempfile import NamedTemporaryFile

from lsprotocol.types import (TextDocumentContentChangeEvent_Type1,
                              TextDocumentDidChangeNotification,
                              TextDocumentDidOpenNotification,
                              WorkspaceDidRenameFilesNotification)

from llanguage_test_client.lsp_test_client import OutgoingEvent
from lfortran_language_server.lfortran_lsp_test_client import LFortranLspTestClient


def test_document_manipulation(client: LFortranLspTestClient):
    with NamedTemporaryFile(
            prefix="test_document_manipulation-",
            suffix=".f90",
            delete=True
    ) as tmp_file_1:
        with NamedTemporaryFile(
                prefix="test_document_manipulation-",
                suffix=".f90",
                delete=True
        ) as tmp_file_2:
            doc = client.new_document("fortran")
            doc.write("module module_function_call1\n")
            doc.write("end module module_function_call1\n")
            doc.save(tmp_file_1.name)

            assert os.path.exists(tmp_file_1.name)

            def sent_text_document_did_open(event: OutgoingEvent) -> bool:
                if isinstance(event.data, TextDocumentDidOpenNotification):
                    notification = event.data
                    return notification.params.text_document.uri == doc.uri
                return False
            assert client.has_outgoing_event(sent_text_document_did_open)

            assert client.await_validation(doc.uri, doc.version) is not None

            rdoc = client.get_remote_document(doc.uri)
            assert doc.uri == rdoc["uri"]
            assert doc.version == rdoc["version"]

            with open(tmp_file_1.name) as f:
                assert doc.text == rdoc["text"] == f.read() == "\n".join([
                    "module module_function_call1",
                    "end module module_function_call1",
                ]) + "\n"

            doc.cursor = 1,6
            doc.write("foo")
            doc.save()

            def sent_text_document_did_write_foo(event: OutgoingEvent) -> bool:
                if isinstance(event.data, TextDocumentDidChangeNotification):
                    notification = event.data
                    params = notification.params
                    if params.text_document.uri == doc.uri:
                        content_changes = params.content_changes
                        if len(content_changes) == 1:
                            content_change = content_changes[0]
                            if isinstance(content_change,
                                          TextDocumentContentChangeEvent_Type1):
                                range = content_change.range
                                start = range.start
                                end = range.end
                                return start.line == 0 \
                                    and start.character == 5 \
                                    and end.line == 0 \
                                    and end.character == 5 \
                                    and content_change.text == "foo"
                return False
            assert client.has_outgoing_event(sent_text_document_did_write_foo)

            assert client.await_validation(doc.uri, doc.version) is not None

            rdoc = client.get_remote_document(doc.uri)
            assert doc.uri == rdoc["uri"]
            assert doc.version == rdoc["version"]

            with open(tmp_file_1.name) as f:
                assert doc.text == rdoc["text"] == f.read() == "\n".join([
                    "modulfooe module_function_call1",
                    "end module module_function_call1",
                ]) + "\n"

            doc.cursor = 1,6
            doc.replace("bar")
            doc.save()

            def sent_text_document_did_replace_foo_with_bar(
                    event: OutgoingEvent
            ) -> bool:
                if isinstance(event.data, TextDocumentDidChangeNotification):
                    notification = event.data
                    params = notification.params
                    if params.text_document.uri == doc.uri:
                        content_changes = params.content_changes
                        if len(content_changes) == 1:
                            content_change = content_changes[0]
                            if isinstance(content_change, TextDocumentContentChangeEvent_Type1):
                                range = content_change.range
                                start = range.start
                                end = range.end
                                return start.line == 0 \
                                    and start.character == 5 \
                                    and end.line == 0 \
                                    and end.character == 8 \
                                    and content_change.text == "bar"
                return False
            assert client.has_outgoing_event(
                sent_text_document_did_replace_foo_with_bar
            )

            assert client.await_validation(doc.uri, doc.version) is not None

            rdoc = client.get_remote_document(doc.uri)
            assert doc.uri == rdoc["uri"]
            assert doc.version == rdoc["version"]

            with open(tmp_file_1.name) as f:
                assert doc.text == rdoc["text"] == f.read() == "\n".join([
                    "modulbare module_function_call1",
                    "end module module_function_call1",
                ]) + "\n"

            doc.backspace()
            doc.save()

            def sent_text_document_did_backspace_r(event: OutgoingEvent) -> bool:
                if isinstance(event.data, TextDocumentDidChangeNotification):
                    notification = event.data
                    params = notification.params
                    if params.text_document.uri == doc.uri:
                        content_changes = params.content_changes
                        if len(content_changes) == 1:
                            content_change = content_changes[0]
                            if isinstance(content_change,
                                          TextDocumentContentChangeEvent_Type1):
                                range = content_change.range
                                start = range.start
                                end = range.end
                                return start.line == 0 \
                                    and start.character == 7 \
                                    and end.line == 0 \
                                    and end.character == 8 \
                                    and content_change.text == ""
                return False
            assert client.has_outgoing_event(sent_text_document_did_backspace_r)

            assert client.await_validation(doc.uri, doc.version) is not None

            rdoc = client.get_remote_document(doc.uri)
            assert doc.uri == rdoc["uri"]
            assert doc.version == rdoc["version"]

            with open(tmp_file_1.name) as f:
                assert doc.text == rdoc["text"] == f.read() == "\n".join([
                    "modulbae module_function_call1",
                    "end module module_function_call1",
                ]) + "\n"

            doc.cursor = 1,6
            doc.delete(2)
            doc.save()

            def sent_text_document_did_delete_ba(event: OutgoingEvent) -> bool:
                if isinstance(event.data, TextDocumentDidChangeNotification):
                    notification = event.data
                    params = notification.params
                    if params.text_document.uri == doc.uri:
                        content_changes = params.content_changes
                        if len(content_changes) == 1:
                            content_change = content_changes[0]
                            if isinstance(content_change,
                                          TextDocumentContentChangeEvent_Type1):
                                range = content_change.range
                                start = range.start
                                end = range.end
                                return start.line == 0 \
                                    and start.character == 5 \
                                    and end.line == 0 \
                                    and end.character == 7 \
                                    and content_change.text == ""
                return False
            assert client.has_outgoing_event(sent_text_document_did_delete_ba)

            assert client.await_validation(doc.uri, doc.version) is not None

            rdoc = client.get_remote_document(doc.uri)
            assert doc.uri == rdoc["uri"]
            assert doc.version == rdoc["version"]

            with open(tmp_file_1.name) as f:
                assert doc.text == rdoc["text"] == f.read() == "\n".join([
                    "module module_function_call1",
                    "end module module_function_call1",
                ]) + "\n"

            doc.move(tmp_file_2.name)

            def sent_workspace_did_rename_file(event: OutgoingEvent) -> bool:
                if isinstance(event.data, WorkspaceDidRenameFilesNotification):
                    notification = event.data
                    for record in notification.params.files:
                        if record.old_uri == f"file://{tmp_file_1.name}" \
                           and record.new_uri == f"file://{tmp_file_2.name}":
                            return True
                return False
            assert client.has_outgoing_event(sent_workspace_did_rename_file)

            assert not os.path.exists(tmp_file_1.name)
            open(tmp_file_1.name, "w").close() # ensure it exists during clean-up
            assert os.path.exists(tmp_file_2.name)

            rdoc = client.get_remote_document(doc.uri)
            assert doc.uri == rdoc["uri"]
            assert doc.version == rdoc["version"]

            with open(tmp_file_2.name) as f:
                assert doc.text == rdoc["text"] == f.read() == "\n".join([
                    "module module_function_call1",
                    "end module module_function_call1",
                ]) + "\n"

            doc.remove()
            assert not os.path.exists(tmp_file_2.name)
            open(tmp_file_2.name, "w").close() # ensure it exists during clean-up



passes = [
        "array_struct_temporary",
        "replace_openmp",
        "replace_function_call_in_declaration",
        "replace_array_passed_in_function_call",
        "replace_array_op",
        "replace_class_constructor",
        "dead_code_removal",
        "replace_div_to_mul",
        "replace_do_loops",
        "replace_flip_sign",
        "replace_fma",
        "replace_for_all",
        "while_else",
        "wrap_global_stmts",
        "replace_implied_do_loops",
        "replace_init_expr",
        "inline_function_calls",
        "replace_symbolic",
        "replace_intrinsic_function",
        "loop_unroll",
        "loop_vectorise",
        "nested_vars",
        "replace_param_to_const",
        "array_by_data",
        "compare",
        "list_expr",
        "replace_print_arr",
        "replace_print_list_tuple",
        "replace_print_struct_type",
        "replace_select_case",
        "replace_sign_from_value",
        "create_subroutine_from_function",
        "transform_optional_argument_functions",
        "unused_functions",
        "update_array_dim_intrinsic_calls",
        "replace_where",
        "unique_symbols",
        "insert_deallocate",
        "promote_allocatable_to_nonallocatable",
        "replace_with_compile_time_values"
]



for name in passes:
    print(f"Processing: {name}")
    name_up = name.upper()
    header = rf"""#ifndef LIBASR_PASS_{name_up}_H
#define LIBASR_PASS_{name_up}_H

#include <libasr/asr.h>
#include <libasr/utils.h>

namespace LCompilers {{

    void pass_{name}(Allocator &al, ASR::TranslationUnit_t &unit,
                                const PassOptions &pass_options);

}} // namespace LCompilers

#endif // LIBASR_PASS_{name_up}_H
"""
    header_filename = f"pass/{name}.h"
    f = open(header_filename, "w")
    f.write(header)
